import pytest
import yaml

from nasap_net import Assembly
from nasap_net.pipelines import concatenate_assemblies_without_isom_checks


def test_basic(tmp_path):
    assemblies_paths = [
        tmp_path / 'assemblies-0.yaml',
        tmp_path / 'assemblies-1.yaml']
    
    ASSEMBLIES = [
        {0: Assembly({'L0': 'L'}), 1: Assembly({'M0': 'M'})},
        {0: Assembly({'L1': 'L'}), 1: Assembly({'M1': 'M'})}]
    
    for path, assems in zip(assemblies_paths, ASSEMBLIES):
        with open(path, 'w') as f:
            yaml.dump(assems, f)

    OUTPUT_PATH = tmp_path / 'concatenated.yaml'

    concatenate_assemblies_without_isom_checks(
        assemblies_path_list=assemblies_paths,
        resulting_assems_path=str(OUTPUT_PATH),
        start=0, overwrite=True, verbose=False)

    with open(OUTPUT_PATH, 'r') as f:
        output_data = yaml.safe_load(f)

    assert output_data == {
        0: Assembly({'L0': 'L'}), 1: Assembly({'M0': 'M'}),
        2: Assembly({'L1': 'L'}), 3: Assembly({'M1': 'M'})}


def test_paths_order_preservation(tmp_path):
    """Test that the order of the paths is preserved
    regardless of their alphabetical order.
    """
    # The order of the paths is different from their alphabetical order.
    assemblies_paths = [tmp_path / 'B.yaml', tmp_path / 'A.yaml']
    
    with open(tmp_path / 'B.yaml', 'w') as f:
        yaml.dump({0: Assembly({'first': 'first'})}, f)
    with open(tmp_path / 'A.yaml', 'w') as f:
        yaml.dump({0: Assembly({'second': 'second'})}, f)
    
    output_path = tmp_path / 'concatenated.yaml'

    concatenate_assemblies_without_isom_checks(
        assemblies_path_list=assemblies_paths,
        resulting_assems_path=str(output_path),
        start=0, overwrite=True, verbose=False)

    with open(output_path, 'r') as f:
        output_data = yaml.safe_load(f)

    # The order of the paths should be preserved.
    assert output_data == {
        0: Assembly({'first': 'first'}), 1: Assembly({'second': 'second'})}


def test_assembly_order_preservation(tmp_path):
    """Test that the order of the assemblies in each file is preserved
    regardless of the alphabetical order of the IDs.
    """
    assemblies_paths = [tmp_path / 'assemblies-0.yaml']
    
    # The order is different from their alphabetical order.
    ASSEMBLIES = [{
        'B': Assembly({'first': 'L'}), 
        'A': Assembly({'second': 'L'}),
        }]
    
    # Save the assemblies preserving the order
    for path, assems in zip(assemblies_paths, ASSEMBLIES):
        with open(path, 'w') as f:
            yaml.dump(assems, f, sort_keys=False)  # keep the order

    OUTPUT_PATH = tmp_path / 'concatenated.yaml'

    concatenate_assemblies_without_isom_checks(
        assemblies_path_list=assemblies_paths,
        resulting_assems_path=str(OUTPUT_PATH),
        start=0, overwrite=True, verbose=False)

    with open(OUTPUT_PATH, 'r') as f:
        output_data = yaml.safe_load(f)

    # The order of the assemblies should be preserved
    assert output_data == {
        0: Assembly({'first': 'L'}), 1: Assembly({'second': 'L'})}


def test_start(tmp_path):
    ASSEMBLIES = {
        0: Assembly({'L0': 'L'}),
        1: Assembly({'M0': 'M'}),
    }

    filepath = tmp_path / 'assemblies.yaml'
    with open(filepath, 'w') as f:
        yaml.dump(ASSEMBLIES, f)

    OUTPUT_PATH = tmp_path / 'concatenated.yaml'

    concatenate_assemblies_without_isom_checks(
        assemblies_path_list=[filepath],
        resulting_assems_path=str(OUTPUT_PATH),
        start=10, overwrite=True, verbose=False)

    with open(OUTPUT_PATH, 'r') as f:
        output_data = yaml.safe_load(f)

    assert output_data == {
        10: Assembly({'L0': 'L'}),
        11: Assembly({'M0': 'M'})}


def test_overwrite_true(tmp_path):
    ASSEMBLIES = {0: Assembly({'L0': 'L'})}
    filepath = tmp_path / 'assemblies.yaml'
    with open(filepath, 'w') as f:
        yaml.dump(ASSEMBLIES, f)
    
    OUTPUT_PATH = tmp_path / 'concatenated.yaml'
    with open(OUTPUT_PATH, 'w') as f:
        yaml.dump('previous data', f)

    concatenate_assemblies_without_isom_checks(
        assemblies_path_list=[filepath],
        resulting_assems_path=str(OUTPUT_PATH),
        start=0, overwrite=True, verbose=False)
    
    with open(OUTPUT_PATH, 'r') as f:
        output_data = yaml.safe_load(f)

    assert output_data == {0: Assembly({'L0': 'L'})}


def test_overwrite_false(tmp_path):
    ASSEMBLIES = {0: Assembly({'L0': 'L'})}
    filepath = tmp_path / 'assemblies.yaml'
    with open(filepath, 'w') as f:
        yaml.dump(ASSEMBLIES, f)
    
    OUTPUT_PATH = tmp_path / 'concatenated.yaml'
    with open(OUTPUT_PATH, 'w') as f:
        yaml.dump('previous data', f)

    with pytest.raises(FileExistsError):
         concatenate_assemblies_without_isom_checks(
            assemblies_path_list=[filepath],
            resulting_assems_path=str(OUTPUT_PATH),
            start=0, overwrite=False, verbose=False)


def test_verbose_true(tmp_path, capsys):
    ASSEMBLIES = {0: Assembly({'L0': 'L'})}
    filepath = tmp_path / 'assemblies.yaml'
    with open(filepath, 'w') as f:
        yaml.dump(ASSEMBLIES, f)
    
    OUTPUT_PATH = tmp_path / 'concatenated.yaml'

    concatenate_assemblies_without_isom_checks(
        assemblies_path_list=[filepath],
        resulting_assems_path=str(OUTPUT_PATH),
        start=0, overwrite=True, verbose=True)

    captured = capsys.readouterr()
    assert f'Finished reading from "{filepath}"' in captured.out
    assert 'Concatenating assembly lists...' in captured.out
    assert 'Concatenation completed.' in captured.out
    assert f'Successfully saved to "{OUTPUT_PATH}".' in captured.out


def test_verbose_false(tmp_path, capsys):
    ASSEMBLIES = {0: Assembly({'L0': 'L'})}
    filepath = tmp_path / 'assemblies.yaml'
    with open(filepath, 'w') as f:
        yaml.dump(ASSEMBLIES, f)
    
    OUTPUT_PATH = tmp_path / 'concatenated.yaml'

    concatenate_assemblies_without_isom_checks(
        assemblies_path_list=[filepath],
        resulting_assems_path=str(OUTPUT_PATH),
        start=0, overwrite=True, verbose=False)

    captured = capsys.readouterr()
    assert captured.out == ''
    assert captured.err == ''


if __name__ == '__main__':
    pytest.main(['-vv', __file__])



from frozendict import frozendict

from nasap_net.isomorphism import Isomorphism, get_all_isomorphisms
from nasap_net.models import Assembly, BindingSite, Bond, Component


def test_get_all_isomorphisms():
    M = Component(kind='M', sites=[0, 1])
    X = Component(kind='X', sites=[0])
    MX2 = Assembly(
        components={'M1': M, 'X1': X, 'X2': X},
        bonds=[Bond('M1', 0, 'X1', 0), Bond('M1', 1, 'X2', 0)]
    )
    MX2_2 = Assembly(
        components={'M10': M, 'X10': X, 'X20': X},
        bonds=[Bond('M10', 0, 'X10', 0), Bond('M10', 1, 'X20', 0)]
    )

    isoms = get_all_isomorphisms(MX2, MX2_2)
    assert len(isoms) == 2
    assert isoms == {
        Isomorphism(
            comp_id_mapping=frozendict(
                {'M1': 'M10', 'X1': 'X10', 'X2': 'X20'}),
            binding_site_mapping=frozendict({
                BindingSite('M1', 0): BindingSite('M10', 0),
                BindingSite('M1', 1): BindingSite('M10', 1),
                BindingSite('X1', 0): BindingSite('X10', 0),
                BindingSite('X2', 0): BindingSite('X20', 0),
            })
        ),
        Isomorphism(
            comp_id_mapping=frozendict(
                {'M1': 'M10', 'X1': 'X20', 'X2': 'X10'}),
            binding_site_mapping=frozendict({
                BindingSite('M1', 0): BindingSite('M10', 1),
                BindingSite('M1', 1): BindingSite('M10', 0),
                BindingSite('X1', 0): BindingSite('X20', 0),
                BindingSite('X2', 0): BindingSite('X10', 0),
            })
        ),
    }



import pytest

from nasap_net.debug_utils import BondsetsComparer

# Sample structure:
# M2L2-ring
# []: bond name
# 
#         M1
# [1] -> /  \ <- [4]
#      L1    L2
# [2] -> \  / <- [3]
#         M2

def test_equal():
    FIRST_BONDSETS = [[1], [1, 2]]
    SECOND_BONDSETS = [[1], [2, 1]]
    comparer = BondsetsComparer(FIRST_BONDSETS, SECOND_BONDSETS)
    assert comparer.are_equal


def test_inequal():
    FIRST_BONDSETS = [[1], [1, 2]]
    SECOND_BONDSETS = [[1], [2, 3]]
    comparer = BondsetsComparer(FIRST_BONDSETS, SECOND_BONDSETS)
    assert not comparer.are_equal


def test_equal_with_sym_ops():
    FIRST_BONDSETS = [[1], [1, 2]]
    SECOND_BONDSETS = [[2], [3, 4]]
    SYM_OPS = {
        'sigma': {1: 2, 2: 1, 3: 4, 4: 3},  # [1] -> [2]
        'C2': {1: 3, 2: 4, 3: 1, 4: 2},  # [1, 2] -> [3, 4]
        }
    comparer = BondsetsComparer(FIRST_BONDSETS, SECOND_BONDSETS, SYM_OPS)
    assert comparer.are_equal


def test_mapping_and_diffs():
    FIRST_BONDSETS = [[1], [1, 2]]
    SECOND_BONDSETS = [[2], [2, 3]]
    SYM_OPS = {
        'sigma': {1: 2, 2: 1, 3: 4, 4: 3},  # [1] -> [2]
        'C2': {1: 3, 2: 4, 3: 1, 4: 2},  # [1, 2] -> [3, 4]
        }
    comparer = BondsetsComparer(FIRST_BONDSETS, SECOND_BONDSETS, SYM_OPS)
    assert comparer.first_to_second == {frozenset([1]): frozenset([2])}
    assert comparer.only_in_first == {frozenset([1, 2])}
    assert comparer.only_in_second == {frozenset([2, 3])}


if __name__ == '__main__':
    pytest.main(['-vv', __file__])



import pytest

from nasap_net import Assembly, AuxEdge, Component, has_aux_edges


def test_has_aux_edges_true():
    MX4 = Assembly(
        {'M1': 'M', 'X1': 'X', 'X2': 'X', 'X3': 'X', 'X4': 'X'},
        [('M1.a', 'X1.a'), ('M1.b', 'X2.a'), ('M1.c', 'X3.a'), ('M1.d', 'X4.a')])
    COMPONENT_STRUCTURES = {
        'M': Component(
            {'a', 'b', 'c', 'd'}, 
            {AuxEdge('a', 'b', 'cis'), AuxEdge('b', 'c', 'cis'),
             AuxEdge('c', 'd', 'cis'), AuxEdge('d', 'a', 'cis')}),
        'X': Component({'a'})}
    assert has_aux_edges(MX4, COMPONENT_STRUCTURES)


def test_has_aux_edges_false():
    MX4 = Assembly(
        {'M1': 'M', 'X1': 'X', 'X2': 'X', 'X3': 'X', 'X4': 'X'},
        [('M1.a', 'X1.a'), ('M1.b', 'X2.a'), ('M1.c', 'X3.a'), ('M1.d', 'X4.a')])
    COMPONENT_STRUCTURES = {
        'M': Component({'a', 'b', 'c', 'd'}),  # No aux edges
        'X': Component({'a'})}
    assert not has_aux_edges(MX4, COMPONENT_STRUCTURES)


def test_has_aux_edges_false2():
    MX4 = Assembly(
        {'M1': 'M', 'X1': 'X', 'X2': 'X', 'X3': 'X', 'X4': 'X'},
        [('M1.a', 'X1.a'), ('M1.b', 'X2.a'), ('M1.c', 'X3.a'), ('M1.d', 'X4.a')])
    COMPONENT_STRUCTURES = {
        'M': Component({'a', 'b', 'c', 'd'}),  # No aux edges
        'L': Component(
            {'a', 'b', 'c', 'd'},
            {AuxEdge('a', 'b', 'cis'), AuxEdge('b', 'c', 'cis'),
             AuxEdge('c', 'd', 'cis'), AuxEdge('d', 'a', 'cis')}),  # Aux edges but not in assembly
        'X': Component({'a'})}
    assert not has_aux_edges(MX4, COMPONENT_STRUCTURES)



if __name__ == '__main__':
    pytest.main(['-vv', __file__])



from collections.abc import Mapping
from functools import reduce


def resolve_chain_map(*maps: Mapping) -> Mapping:
    """
    Resolve a chain of mappings into one.

    Parameters
    ----------
    *maps : Mapping
        Mappings to resolve.

    Returns
    -------
    Mapping
        The resolved mapping.

    Examples
    --------
    >>> input1 = {1: 10, 2: 20}
    >>> input2 = {10: 100, 20: 200}
    >>> input3 = {100: 1000, 200: 2000}
    >>> resolve_chain_map(input1, input2, input3)
    {1: 1000, 2: 2000}

    >>> resolve_chain_map()
    {}

    >>> resolve_chain_map({1: 10})
    {1: 10}
    """
    if not maps:
        return {}

    return reduce(
        lambda x, y: {k: x[v] for k, v in y.items()}, 
        reversed(maps))

    # ==============================
    # Explanation of the logic:

    # Example:
    # m1 = {1: 10, 2: 20}
    # m2 = {10: 100, 20: 200}
    # m3 = {100: 1000, 200: 2000}

    # The following code:
    # ```
    # reduce(
    #     lambda x, y: {k: x[v] for k, v in y.items()},
    #     [m3, m2, m1])
    # ```
    # is equivalent to:
    # ```
    # x = m3  # {100: 1000, 200: 2000}
    # y = m2  # {10: 100, 20: 200}
    # result = {k: x[v] for k, v in y.items()}  # {10: 1000, 20: 2000}
    # x = result  # {10: 1000, 20: 2000}
    # y = m1  # {1: 10, 2: 20}
    # result = {k: x[v] for k, v in y.items()}  # {1: 1000, 2: 2000}
    # ```



from collections.abc import Iterable
from typing import cast

import yaml

from ..aux_edge import AuxEdge
from ..validations import validate_name_of_binding_site
from .bindsite_existence_check import check_bindsites_of_aux_edges_exists

__all__ = ['Component']


class Component(yaml.YAMLObject):
    """A component of an assembly. (Immutable)"""
    yaml_loader = yaml.SafeLoader
    yaml_dumper = yaml.Dumper
    yaml_tag = '!Component'
    yaml_flow_style = None

    def __init__(
            self,
            binding_sites: Iterable[str],
            aux_edges: (
                Iterable[AuxEdge] 
                | Iterable[tuple[str, str, str]] | None
                ) = None):
        """
        Parameters
        ----------
        component_kind : str
            The component type, e.g., 'M', 'L', 'L1', 'X', etc.
        binding_sites : Iterable[str]
            The binding sites. Each binding site should be a string.
        aux_edges : Mapping[tuple[str, str], str], optional
            Mapping of auxiliary edges. The keys are tuples of two binding
            sites, and the values are the auxiliary kinds. The binding sites
            should be in the binding_sites.
            Duplicate pairs of binding sites raise an error regardless of the
            order of the binding sites.
        """
        for binding_site in binding_sites:
            validate_name_of_binding_site(binding_site)
        self._binding_sites = frozenset(binding_sites)

        if aux_edges is None:
            self._aux_edges = frozenset[AuxEdge]()
        elif all(isinstance(edge, AuxEdge) for edge in aux_edges):
            aux_edges = cast(Iterable[AuxEdge], aux_edges)
            self._aux_edges = frozenset(aux_edges)
        else:
            aux_edges = cast(Iterable[tuple[str, str, str]], aux_edges)
            self._aux_edges = frozenset(
                {AuxEdge(*edge) for edge in aux_edges})

        check_bindsites_of_aux_edges_exists(
            self._aux_edges, self._binding_sites)
    
    def __eq__(self, value: object) -> bool:
        if not isinstance(value, Component):
            return False
        return (
            self.binding_sites == value.binding_sites
            and self.aux_edges == value.aux_edges)
    
    def __repr__(self) -> str:
        if not self.aux_edges:
            return f'Component({sorted(self.binding_sites)!r})'
        return (
            f'Component({sorted(self.binding_sites)!r}, '
            f'{sorted(self.aux_edges)!r})'
            )
    
    @property
    def binding_sites(self) -> set[str]:
        return set(self._binding_sites)

    @property
    def aux_edges(self) -> set[AuxEdge]:
        return set(self._aux_edges)
    
    @classmethod
    def from_yaml(cls, loader, node):
        data = loader.construct_mapping(node, deep=True)
        binding_sites = data['binding_sites']
        aux_edges = data.get('aux_edges', None)
        return cls(binding_sites, aux_edges)

    @classmethod
    def to_yaml(cls, dumper, data):
        data_dict = {'binding_sites': sorted(data.binding_sites)}
        if data.aux_edges:
            data_dict['aux_edges'] = sorted(
                data.aux_edges,
                key=lambda edge: sorted(edge.binding_sites))
        return dumper.represent_mapping(
            cls.yaml_tag, data_dict,
            flow_style=cls.yaml_flow_style)



import itertools
from collections.abc import Iterable, Iterator, Mapping

from nasap_net import (Assembly, Component, find_isomorphic_assembly,
                       perform_inter_exchange)
from nasap_net.classes.reaction import InterReaction

from .lib import (compute_unique_bindsites_or_bindsite_sets,
                  enum_valid_entering_bindsites, enum_valid_ml_pairs)

__all__ = ['explore_inter_reactions']


def explore_inter_reactions(
        init_assem_id: int, entering_assem_id: int,
        metal_kind: str, leaving_kind: str, entering_kind: str,
        id_to_assembly: Mapping[int, Assembly],
        hash_to_ids: Mapping[str, Iterable[int]],
        component_structures: Mapping[str, Component],
        ) -> Iterator[InterReaction]:
    init_assem = id_to_assembly[init_assem_id]
    entering_assem = id_to_assembly[entering_assem_id]

    valid_ml_pairs = enum_valid_ml_pairs(
        init_assem_id, init_assem,
        metal_kind, leaving_kind,
        component_structures)
    valid_entering_bindsites = enum_valid_entering_bindsites(
        entering_assem_id, entering_assem,
        entering_kind, component_structures)

    unique_ml_pairs_and_dup_cnts = compute_unique_bindsites_or_bindsite_sets(
        init_assem_id, init_assem, valid_ml_pairs, component_structures)
    unique_entering_bindsites_and_dup_cnts = compute_unique_bindsites_or_bindsite_sets(
        entering_assem_id, entering_assem, valid_entering_bindsites,
        component_structures)

    for ml_pair_and_dup_cnt, entering_bs_and_dup_cnt in itertools.product(
            unique_ml_pairs_and_dup_cnts, unique_entering_bindsites_and_dup_cnts):
        (metal_bs, leaving_bs), ml_pair_dup_cnt = ml_pair_and_dup_cnt
        entering_bs, entering_bs_dup_cnt = entering_bs_and_dup_cnt

        duplicate_count = ml_pair_dup_cnt * entering_bs_dup_cnt
        if init_assem_id == entering_assem_id:
            duplicate_count *= 2

        product, leaving = perform_inter_exchange(
            init_assem, entering_assem, metal_bs, leaving_bs, entering_bs)

        product_id = find_isomorphic_assembly(
            product, id_to_assembly, hash_to_ids, component_structures)

        if product_id is None:
            continue

        if leaving is None:
            yield InterReaction(
                init_assem_id, entering_assem_id, product_id,
                None, metal_bs, leaving_bs, entering_bs,
                duplicate_count)
            continue

        leaving_id = find_isomorphic_assembly(
            leaving, id_to_assembly, hash_to_ids,
            component_structures)
        if leaving_id is None:
            continue

        yield InterReaction(
            init_assem_id, entering_assem_id, product_id, leaving_id,
            metal_bs, leaving_bs, entering_bs,
            duplicate_count)



from nasap_net import Assembly, BindsiteIdConverter
from nasap_net.algorithms.assembly_separation import separate_product_if_possible
from nasap_net.algorithms.union import union_assemblies


def perform_inter_exchange(
        init_assem: Assembly, entering_assem: Assembly,
        metal_bs: str, leaving_bs: str, entering_bs: str,
        ) -> tuple[Assembly, Assembly | None]:
    """Perform an inter-molecular ligand exchange reaction.

    This function returns the resulting assembly(ies) after performing
    a ligand exchange reaction between two assemblies.

    The ligand exchange process can be summarized as follows:
    - The bond between the metal center (metal_bs) and the leaving ligand
        (leaving_bs) in init_assem is broken.
    - A new bond is formed between the metal center (metal_bs) and the
        entering ligand (entering_bs) before or after the bond breaking,
        depending on the specific mechanism (dissociative or associative).

    Parameters
    ----------
    init_assem : Assembly
        The assembly containing the metal center and the leaving ligand.
    entering_assem : Assembly
        The assembly containing the entering ligand.
    metal_bs : str
        The binding site ID of the metal center (on init_assem).
    leaving_bs : str
        The binding site ID of the leaving ligand (on init_assem).
        This binding site should be bonded to metal_bs in init_assem.
    entering_bs : str
        The binding site ID of the entering ligand (on entering_assem).
        This binding site should not be bonded to any other binding site
        in entering_assem.

    Returns
    -------
    main_assembly : Assembly
        The assembly containing the metal center and the entering ligand.
        The leaving ligand may be part of this assembly or may be
        contained in a separated assembly (leaving_assembly).
    leaving_assembly : Assembly or None
        The separated leaving assembly, if it exists; otherwise None.

    Notes
    -----
    - This function renames component IDs in the input assemblies to
      avoid ID conflicts when merging them, by prefixing them with
      'init_' and 'entering_' respectively.
    """
    init_relabel = {
        comp_id: f'init_{comp_id}'
        for comp_id in init_assem.component_ids}
    init_assem = init_assem.rename_component_ids(init_relabel)

    entering_relabel = {
        comp_id: f'entering_{comp_id}'
        for comp_id in entering_assem.component_ids}
    entering_assem = entering_assem.rename_component_ids(entering_relabel)

    metal_bs = f'init_{metal_bs}'
    leaving_bs = f'init_{leaving_bs}'
    entering_bs = f'entering_{entering_bs}'

    init_assem = union_assemblies(init_assem, entering_assem)

    init_assem.remove_bond(metal_bs, leaving_bs)
    init_assem.add_bond(entering_bs, metal_bs)

    # Separate the leaving assembly if possible
    id_converter = BindsiteIdConverter()
    metal_comp, rel_bindsite = id_converter.global_to_local(metal_bs)
    main_assem, leaving_assem = separate_product_if_possible(
        init_assem, metal_comp)
    return main_assem, leaving_assem



from collections.abc import Iterable

from nasap_net import Assembly


def bond_induced_sub_assembly(
        assembly: Assembly, bonds: Iterable[Iterable[str]]) -> Assembly:
    """Create a sub-assembly induced by the bonds.

    The induced sub-assembly contains all the bonds in the `bonds` and
    all the components connected by any bond in the `bonds`.
    """
    component_ids: set[str] = set()
    for bond in bonds:
        component_ids.update(assembly.bond_to_rough_bond(bond))

    components = {
        comp_id: assembly.get_component_kind(comp_id)
        for comp_id in component_ids}

    return Assembly(components, bonds)


def component_induced_sub_assembly(
        assembly: Assembly, component_ids: set[str]) -> Assembly:
    """Create a sub-assembly induced by the components.

    The induced sub-assembly contains all the components in the 
    `component_ids` and all the bonds between the components in the 
    `component_ids`.
    """
    components = {
        comp_id: assembly.get_component_kind(comp_id) for comp_id in component_ids}
    bonds = {
        bond for bond in assembly.bonds
        if set(assembly.bond_to_rough_bond(bond)) <= component_ids}
    # i.e., if the both binding sites of a bond are in the component_ids

    return Assembly(components, bonds)



import pytest

from nasap_net import Assembly
from nasap_net.reaction_classification.utils import find_shortest_cycle
from nasap_net.utils import are_same_circular_perm


def test_no_ring():
    # M1-L1 (linear)
    ML_NO_RING = Assembly({'M1': 'M', 'L1': 'L'})
    assert find_shortest_cycle(ML_NO_RING) == None


def test_one_ring():
    # M1---L1
    # |    |
    # L2---M2

    # One cycle: 
    # 1. M1-L1-M2-L2-(M1)  (length 4)

    M2L2_RING = Assembly(
        {'M1': 'M', 'M2': 'M', 'L1': 'L', 'L2': 'L'},
        [('M1.b', 'L1.a'), ('L1.b', 'M2.a'),
         ('M2.b', 'L2.a'), ('L2.b', 'M1.a')])
    result = find_shortest_cycle(M2L2_RING)
    assert result is not None
    assert are_same_circular_perm(
        result, ['M1', 'L1', 'M2', 'L2'], consider_reverse=True)


def test_one_parallel():
    # Parallel bonds are also considered as a cycle.

    #   M1
    #  /  \  <- parallel bonds
    #  \  /
    #   L1

    # One cycle:
    # 1. M1-L1-(M1)  (length 2, parallel-bond pair)

    ML_RING = Assembly({'M1': 'M', 'L1': 'L'},
                       [('M1.a', 'L1.a'), ('M1.b', 'L1.b')])
    
    result = find_shortest_cycle(ML_RING)
    assert result is not None
    assert are_same_circular_perm(
        result, ['M1', 'L1'], consider_reverse=True)


def test_multi_rings_of_same_length():
    #   -- M1--
    #  /   |   \
    # L1   L2   L3
    #  \   |   /
    #   -- M2--

    # Three cycles with the same length 4:
    # 1. M1-L1-M2-L2-(M1)  (length 4)
    # 2. M1-L2-M2-L3-(M1)  (length 4)
    # 3. M1-L3-M2-L1-(M1)  (length 4)
    M2L3_CAGE = Assembly(
        {'M1': 'M', 'M2': 'M', 'L1': 'L', 'L2': 'L', 'L3': 'L'},
        [('M1.a', 'L1.a'), ('M1.b', 'L2.a'), ('M1.c', 'L3.a'),
         ('M2.a', 'L1.b'), ('M2.b', 'L2.b'), ('M2.c', 'L3.b')])

    result = find_shortest_cycle(M2L3_CAGE)

    assert result is not None
    # Which cycle is returned is not guaranteed.
    assert are_same_circular_perm(
        result, ['M1', 'L1', 'M2', 'L2'], consider_reverse=True)\
        or are_same_circular_perm(
            result, ['M1', 'L2', 'M2', 'L3'], consider_reverse=True)\
        or are_same_circular_perm(
            result, ['M1', 'L3', 'M2', 'L1'], consider_reverse=True)


def test_multi_rings_of_different_length():
    #  -- M1----L3
    # /   |     |
    # L1  L2    M3
    # \   |     |
    #  -- M2----L4

    # Three cycles with different lengths:
    # 1. M1-L1-M2-L2-(M1)  (length 4)
    # 2. M1-L1-M2-L4-M3-L3-(M1)  (length 6)
    # 3. M1-L2-M2-L4-M3-L3-(M1)  (length 6)
    M3L3_RING = Assembly(
        {'M1': 'M', 'M2': 'M', 'M3': 'M',
         'L1': 'L', 'L2': 'L', 'L3': 'L', 'L4': 'L'},
        [('M1.a', 'L1.a'), ('M1.b', 'L2.a'),
         ('M2.a', 'L1.b'), ('M2.b', 'L2.b'),
         ('M1.c', 'L3.a'), ('L3.b', 'M3.a'),
         ('M3.b', 'L4.a'), ('L4.b', 'M2.c'),
         ])

    result = find_shortest_cycle(M3L3_RING)

    assert result is not None
    assert are_same_circular_perm(
        result, ['M1', 'L1', 'M2', 'L2'], consider_reverse=True)


def test_one_ring_and_one_parallel():
    #  L2----M1
    #  |    /  \  <- parallel bonds
    #  |    \  /
    #  M2----L1

    # Three cycles:
    # 1. M1-L1-(M1)  (length 2, parallel-bond pair)
    # 2. M1-L2-M2-L1-(M1)  (length 4, using one of the parallel bonds)
    # 3. M2-L2-M1-L1-(M2)  (length 4, using the other parallel bond)
    ML_PARALLEL_RING = Assembly(
        {'M1': 'M', 'M2': 'M', 'L1': 'L', 'L2': 'L'},
        [('M1.a', 'L1.a'), ('M1.b', 'L1.b'),  # parallel bonds
         ('M1.c', 'L2.a'), ('L2.b', 'M2.a'),
         ('M2.b', 'L1.c')])

    result = find_shortest_cycle(ML_PARALLEL_RING)

    assert result is not None
    assert are_same_circular_perm(
        result, ['M1', 'L1'], consider_reverse=True)


if __name__ == '__main__':
    pytest.main(['-vv', __file__])



from typing import TypeVar

T = TypeVar('T')


def are_same_circular_perm(
        perm1: list[T], perm2: list[T],
        consider_reverse: bool = False) -> bool:
    """Check if two circular permutations are the same.

    Two circular permutations are the same if they are the same
    when rotated. If consider_reverse is True, the function will
    also check if the second permutation is the reverse of some
    rotation of the first permutation.

    For example, the circular permutations [1, 2, 3, 4] and [3, 2, 1, 4]
    are considered the same if consider_reverse is True, but not
    if consider_reverse is False.

    Parameters
    ----------
    perm1 : list[T]
        The first circular permutation.
    perm2 : list[T]
        The second circular permutation.
    consider_reverse : bool, optional
        Whether to consider the reverse of the second permutation,
        by default False.
    
    Returns
    -------
    bool
        True if the two circular permutations are the same.
    """
    if len(perm1) != len(perm2):
        return False

    for i in range(len(perm1)):
        rotated_perm2 = perm2[i:] + perm2[:i]
        if perm1 == rotated_perm2:
            return True
        if consider_reverse and perm1 == rotated_perm2[::-1]:
            return True

    return False



from collections.abc import Hashable, Iterable, Mapping
from itertools import combinations
from typing import TypeVar, cast

from networkx.utils import UnionFind

from nasap_net import Assembly, Component, is_isomorphic

from .grouping_by_hash import group_assemblies_by_hash

__all__ = ['group_assemblies_by_isomorphism']

_T = TypeVar('_T', bound=Hashable)

def group_assemblies_by_isomorphism(
        id_to_assembly: Mapping[_T, Assembly],
        component_structures: Mapping[str, Component],
        *,
        non_isomorphic_groups: Iterable[set[_T]] | None = None
        ) -> dict[_T, set[_T]]:
    """Group duplicates by assembly isomorphism.

    Parameters
    ----------
    id_to_assembly : Mapping[_T, Assembly]
        A mapping from IDs to assemblies.
    component_structures : Mapping[str, Component]
        A mapping from component kinds to component structures.
    non_isomorphic_groups : Iterable[set[_T]], optional
        Predefined groups of assemblies that are guaranteed to be unique
        and non-isomorphic. Assemblies in the same group will not be 
        checked for isomorphism. If None, all assemblies are checked for 
        isomorphism.

    Returns
    -------
    dict[_T, set[_T]]
        A mapping from unique IDs to sets of duplicate IDs.
        Unique IDs are the minimum IDs in each group.
        Duplicate IDs include the unique IDs themselves.

    Notes
    -----
    The parameter `non_isomorphic_groups` is useful when there are
    assemblies that are known to be unique and non-isomorphic. This
    can be used to reduce the number of isomorphism checks, which can
    be computationally expensive.
    """
    hash_to_ids = group_assemblies_by_hash(id_to_assembly, component_structures)
    
    if non_isomorphic_groups is not None:
        id_to_non_isomorphic_group = {
            id_: group for group in non_isomorphic_groups for id_ in group}

    uf = UnionFind(id_to_assembly.keys())

    for hash_, same_hash_ids in hash_to_ids.items():
        if len(same_hash_ids) == 1:
            continue

        for id1, id2 in combinations(same_hash_ids, 2):
            if uf[id1] == uf[id2]:  # Already checked to be isomorphic
                continue
            if non_isomorphic_groups is not None:
                non_isom_group1 = id_to_non_isomorphic_group.get(id1)
                non_isom_group2 = id_to_non_isomorphic_group.get(id2)
                if non_isom_group1 is not None and non_isom_group1 == non_isom_group2:
                    # Already guaranteed to be non-isomorphic
                    continue
            if is_isomorphic(  # Need isomorphism check
                    id_to_assembly[id1], id_to_assembly[id2],
                    component_structures):
                uf.union(id1, id2)

    grouped_ids = cast(Iterable[set[_T]], uf.to_sets())
    unique_id_to_ids = {min(ids): ids for ids in grouped_ids}
    return unique_id_to_ids



from collections.abc import Iterable, Mapping

from nasap_net import Assembly, Component
from nasap_net.algorithms.hashing import calc_graph_hash_of_assembly
from nasap_net.algorithms.isomorphism import is_isomorphic


def find_isomorphic_assembly(
        target_assembly: Assembly,
        id_to_assembly: Mapping[int, Assembly],
        hash_to_ids: Mapping[str, Iterable[int]],
        component_structures: Mapping[str, Component]
        ) -> int | None:
    # TODO: Reduce the number of calls to calc_wl_hash_of_assembly.
    # Maybe we can use the numbers of each component kind in the assembly.
    hash_ = calc_graph_hash_of_assembly(target_assembly, component_structures)
    if hash_ not in hash_to_ids:
        return None
    for candidate_id in hash_to_ids[hash_]:
        candidate = id_to_assembly[candidate_id]
        if is_isomorphic(
                target_assembly, candidate, component_structures):
            return candidate_id
    return None



from .assembly import Assembly
from .aux_edge import AuxEdge
from .bindsite_id_converter import BindsiteIdConverter
from .component import Component
from .reaction import (InterOrIntra, InterReaction, IntraReaction,
                       ReactionBase)
from .rich_reaction import InterReactionRich, IntraReactionRich, \
    RichReactionBase



from collections.abc import Hashable, Iterable, Mapping
from typing import TypeVar

from .is_new_check import is_new_under_symmetry

__all__ = ['enum_single_bond_subsets']

_T = TypeVar('_T', bound=Hashable)


def enum_single_bond_subsets(
        bonds: Iterable[_T],
        sym_ops: Mapping[str, Mapping[_T, _T]] | None = None
        ) -> set[frozenset[_T]]:
    """Enumerate single-bond subsets of bonds 
    excluding disconnected ones and symmetry-equivalent ones.
    """
    found: set[frozenset[_T]] = set()

    # NOTE: The order of the iteration should be fixed to make the
    # result deterministic.
    for bond in bonds:
        assembly = {bond}
        if is_new_under_symmetry(found, assembly, sym_ops):
            found.add(frozenset(assembly))

    return set(found)



from dataclasses import dataclass

from nasap_net import Assembly, perform_inter_exchange, perform_intra_exchange


@dataclass(frozen=True)
class _ReactionResult:
    product_assembly: Assembly
    leaving_assembly: Assembly | None
    metal_bs: str
    leaving_bs: str
    entering_bs: str


def _generate_reaction_result(
        init_assembly: Assembly, entering_assembly: Assembly | None,
        metal_bs: str, leaving_bs: str, entering_bs: str
        ) -> _ReactionResult:
    if entering_assembly is None:  # intra
        product, leaving = perform_intra_exchange(
            init_assembly, metal_bs, leaving_bs, entering_bs)
        return _ReactionResult(
            product, leaving, metal_bs, leaving_bs, entering_bs)
    else:
        product, leaving = perform_inter_exchange(
            init_assembly, entering_assembly,
            metal_bs, leaving_bs, entering_bs)
        # Note that the `perform_inter_exchange` renames binding sites.
        return _ReactionResult(
            product, leaving,
            f'init_{metal_bs}',
            f'init_{leaving_bs}',
            f'entering_{entering_bs}',
            )



import pytest
import yaml

from nasap_net import Assembly
from nasap_net.pipelines import concatenate_assemblies_pipeline


def test_basic(tmp_path):
    assemblies_paths = [
        tmp_path / 'first.yaml',
        tmp_path / 'second.yaml']
    
    ASSEMBLIES = [
        {0: Assembly({'first': 'L'}), 1: Assembly({'first': 'M'})},
        {0: Assembly({'second': 'L'}), 1: Assembly({'second': 'M'})}]
    
    for path, assems in zip(assemblies_paths, ASSEMBLIES):
        with open(path, 'w') as f:
            yaml.dump(assems, f)

    components_path = tmp_path / 'components.yaml'
    with open(components_path, 'w') as f:
        yaml.dump({'component_kinds': {}}, f)

    OUTPUT_PATH = tmp_path / 'concatenated.yaml'

    concatenate_assemblies_pipeline(
        assemblies_path_list=assemblies_paths,
        components_path=components_path,
        resulting_assems_path=str(OUTPUT_PATH),
        start=0, overwrite=True, verbose=False)

    with open(OUTPUT_PATH, 'r') as f:
        output_data = yaml.safe_load(f)

    assert output_data == {
        0: Assembly({'first': 'L'}), 1: Assembly({'first': 'M'})}
    # The second assemblies are not included because they are isomorphic
    # to the first ones.


def test_duplicates_within_files(tmp_path):
    assemblies_paths = [tmp_path / 'first.yaml']
    
    ASSEMBLIES = [
        {0: Assembly({'first': 'L'}), 1: Assembly({'second': 'L'})}]
    
    for path, assems in zip(assemblies_paths, ASSEMBLIES):
        with open(path, 'w') as f:
            yaml.dump(assems, f)

    components_path = tmp_path / 'components.yaml'
    with open(components_path, 'w') as f:
        yaml.dump({'component_kinds': {}}, f)

    OUTPUT_PATH = tmp_path / 'concatenated.yaml'

    concatenate_assemblies_pipeline(
        assemblies_path_list=assemblies_paths,
        components_path=components_path,
        resulting_assems_path=str(OUTPUT_PATH),
        start=0, overwrite=True, verbose=False)

    with open(OUTPUT_PATH, 'r') as f:
        output_data = yaml.safe_load(f)

    # The second assembly is not included because it is isomorphic to 
    # the first one.
    assert output_data == {0: Assembly({'first': 'L'})}


def test_already_unique_within_files(tmp_path):
    assemblies_paths = [tmp_path / 'first.yaml']
    
    ASSEMBLIES = [
        {0: Assembly({'first': 'L'}), 1: Assembly({'second': 'L'})}]
    
    for path, assems in zip(assemblies_paths, ASSEMBLIES):
        with open(path, 'w') as f:
            yaml.dump(assems, f)

    components_path = tmp_path / 'components.yaml'
    with open(components_path, 'w') as f:
        yaml.dump({'component_kinds': {}}, f)

    OUTPUT_PATH = tmp_path / 'concatenated.yaml'

    concatenate_assemblies_pipeline(
        assemblies_path_list=assemblies_paths,
        components_path=components_path,
        resulting_assems_path=str(OUTPUT_PATH),
        already_unique_within_files=True,  # This is the only difference
        start=0, overwrite=True, verbose=False)

    with open(OUTPUT_PATH, 'r') as f:
        output_data = yaml.safe_load(f)

    # The second assembly is included even though it is isomorphic to
    # the first one. This is because the parameter 
    # `already_unique_within_files` is set to True, which means that
    # the isomorphism check is skipped for the assemblies within the
    # same file.
    assert output_data == {
        0: Assembly({'first': 'L'}), 1: Assembly({'second': 'L'})}


def test_skip_isomorphism_checks(tmp_path):
    assemblies_paths = [
        tmp_path / 'first.yaml',
        tmp_path / 'second.yaml']
    
    ASSEMBLIES = [
        {0: Assembly({'first': 'L'})},
        {0: Assembly({'second': 'L'})}]  # isomorphic to the first one
    
    for path, assems in zip(assemblies_paths, ASSEMBLIES):
        with open(path, 'w') as f:
            yaml.dump(assems, f)

    components_path = tmp_path / 'components.yaml'
    with open(components_path, 'w') as f:
        yaml.dump({'component_kinds': {}}, f)

    OUTPUT_PATH = tmp_path / 'concatenated.yaml'

    concatenate_assemblies_pipeline(
        assemblies_path_list=assemblies_paths,
        components_path=components_path,
        resulting_assems_path=str(OUTPUT_PATH),
        skip_isomorphism_checks=True,  # This is the only difference
        start=0, overwrite=True, verbose=False)
    
    with open(OUTPUT_PATH, 'r') as f:
        output_data = yaml.safe_load(f)

    # The second assembly is included even though it is isomorphic to
    # the first one. This is because the parameter `skip_isomorphism_checks`
    # is set to True, which means that the isomorphism check is skipped
    # for all assemblies.

    assert output_data == {
        0: Assembly({'first': 'L'}), 1: Assembly({'second': 'L'})}


if __name__ == '__main__':
    pytest.main(['-vv', __file__])



from collections.abc import Iterable, Mapping
from dataclasses import dataclass
from typing import TypeVar

from nasap_net.types import ID
from ..assembly import Assembly
from ..component import Component
from ..light_assembly import LightAssembly


@dataclass(frozen=True)
class ConversionResult:
    light_assemblies: Mapping[ID, LightAssembly]
    components: Mapping[str, Component]


class InconsistentComponentKindError(Exception):
    """Raised when there are inconsistent definitions for a component kind,
    i.e., the same kind name corresponds to different component structures.
    """
    def __init__(
            self,
            component_kind: str,
            message: str = 'Inconsistent definitions for component kind.'
    ) -> None:
        self.component_kind = component_kind
        super().__init__(f'{message}: Kind: "{component_kind}".')


_T = TypeVar('_T', bound=ID)

def convert_assemblies_to_light_ones(
        assemblies: Mapping[_T, Assembly],
) -> ConversionResult:
    light_assemblies = _assemblies_to_light_assemblies(assemblies)

    try:
        components = _extract_components(assemblies.values())
    except InconsistentComponentKindError as e:
        raise InconsistentComponentKindError(e.component_kind) from None

    return ConversionResult(
        light_assemblies=light_assemblies,
        components=components,
    )


def _extract_components(
        assemblies: Iterable[Assembly],
) -> dict[str, Component]:
    components: dict[str, Component] = {}
    for assembly in assemblies:
        for comp in assembly.components.values():
            if comp.kind in components:
                if comp != components[comp.kind]:
                    raise InconsistentComponentKindError(comp.kind)
            else:
                components[comp.kind] = comp
    return components


def _assemblies_to_light_assemblies(
        assemblies: Mapping[_T, Assembly],
        ) -> Mapping[ID, LightAssembly]:
    return {
        assembly_id: LightAssembly.from_assembly(assembly)
        for assembly_id, assembly in assemblies.items()
    }



from collections.abc import Mapping
from itertools import product

from cachetools import cached
from cachetools.keys import hashkey

from nasap_net import Assembly, Component

_cache = {}  # type: ignore

def clear_cache_for_enum_valid_mles_for_intra():
    """Clear the cache used by `enum_valid_mles_for_intra`."""
    _cache.clear()


def _cache_key(
        assembly_id: str | int,
        assembly: Assembly,
        metal_kind: str, leaving_kind: str, entering_kind: str,
        component_structures: Mapping[str, Component],
        ) -> tuple:
    return hashkey(assembly_id, metal_kind, leaving_kind, entering_kind)


@cached(cache=_cache, key=_cache_key)
def enum_valid_mles_for_intra(
        assembly_id: str | int,  # only used for caching
        assembly: Assembly,
        metal_kind: str, leaving_kind: str, entering_kind: str,
        component_structures: Mapping[str, Component],
        ) -> list[tuple[str, str, str]]:
    """
    Get all possible metal-leaving-entering (MLE) site trios in an assembly.

    This function identifies all valid combinations of metal bindsites, 
    leaving bindsites, and entering bindsites based on the specified 
    component kinds and their connectivity.

    Returned trios meet the following conditions:
    - The metal bindsite and leaving bindsite are connected to each other.
    - The component kind of the metal bindsite is metal_kind.
    - The component kind of the leaving bindsite is leaving_kind.
    - The entering bindsite is free and has the component kind entering_kind.

    Parameters
    ----------
    assembly : Assembly
        The assembly containing the bindsites.
    component_structures : Mapping[str, ComponentStructure]
        A mapping of component names to their structures.
    metal_kind : str
        The kind of the metal bindsite.
    leaving_kind : str
        The kind of the leaving bindsite.
    entering_kind : str
        The kind of the entering bindsite.

    Returns
    -------
    list of tuple of str
        A list of tuples, each containing a metal bindsite, a leaving bindsite, 
        and an entering bindsite.

    Note
    -----
    The result is cached to avoid redundant computation.
    The key for caching is based on the assembly_id, metal_kind, leaving_kind,
    and entering_kind, i.e., assembly and component_structures are not used.
    Make sure to provide the same assembly_id for the same assembly, and not to
    change the component_structures between calls.
    """
    # Pairs of metal bindsites and leaving bindsites
    # that meet the following conditions:
    # - The bindsites are connected to each other
    # - The component kind of the metal bindsite is metal_kind
    # - The component kind of the leaving bindsite is leaving_kind
    valid_ml_pairs = []
    for metal_bs in assembly.get_all_bindsites_of_kind(
            metal_kind, component_structures):
        connected_bs = assembly.get_connected_bindsite(metal_bs)
        if connected_bs is None:  # metal bindsite is free
            continue
        if assembly.get_component_kind_of_bindsite(
                connected_bs) == leaving_kind:
            valid_ml_pairs.append((metal_bs, connected_bs))

    # Entering bindsites that meet the following conditions:
    # - The bindsite is free
    # - The component kind of the entering bindsite is entering_kind
    entering_bindsites = [
        bs for bs in assembly.get_all_bindsites_of_kind(
        entering_kind, component_structures)
        if assembly.is_free_bindsite(bs)]

    # Trios of metal bindsites, leaving bindsites, and entering bindsites.
    return [
        (metal_bs, leaving_bs, entering_bs)
        for (metal_bs, leaving_bs), entering_bs
        in product(valid_ml_pairs, entering_bindsites)]



from copy import deepcopy

import pytest

from nasap_net import Assembly, AuxEdge, Component
from nasap_net.algorithms.isomorphism import pure_is_isomorphic


@pytest.fixture
def assem_without_bonds() -> Assembly:
    assem = Assembly()
    assem.add_component('M1', 'M')
    assem.add_components([('L1', 'L'), ('L2', 'L')])
    assem.add_components([('X1', 'X'), ('X2', 'X')])
    return assem


@pytest.fixture
def component_structures() -> dict[str, Component]:
    M_COMP = Component(
        {'a', 'b', 'c', 'd'},
        {
            AuxEdge('a', 'b', 'cis'), AuxEdge('b', 'c', 'cis'),
            AuxEdge('c', 'd', 'cis'), AuxEdge('d', 'a', 'cis'),})
    L_COMP = Component({'a', 'b'})
    X_COMP = Component({'a'})
    return {'M': M_COMP, 'L': L_COMP, 'X': X_COMP}


@pytest.fixture
def assem1(assem_without_bonds: Assembly) -> Assembly:
    assem1_ = deepcopy(assem_without_bonds)
    assem1_.add_bonds([
        ('M1.a', 'L1.a'), ('M1.b', 'L2.a'),  # cis
        ('M1.c', 'X1.a'), ('M1.d', 'X2.a')])
    return assem1_


@pytest.fixture
def assem2(assem_without_bonds: Assembly) -> Assembly:
    assem2_ = deepcopy(assem_without_bonds)
    assem2_.add_bonds([
        ('M1.a', 'L1.a'), ('M1.c', 'L2.a'),  # trans
        ('M1.b', 'X1.a'), ('M1.d', 'X2.a')])
    return assem2_


def test_pure_is_isomorphic_with_isomorphic_assemblies(
        assem1: Assembly, component_structures: dict[str, Component]
        ) -> None:
    assem3 = deepcopy(assem1)
    assert pure_is_isomorphic(assem1, assem3, component_structures)


def test_pure_is_isomorphic_with_clearly_non_isomorphic_assemblies(
        assem1: Assembly, component_structures: dict[str, Component]):
    assem2 = deepcopy(assem1)
    assem2.remove_bond('M1.a', 'L1.a')
    assert not pure_is_isomorphic(assem1, assem2, component_structures)
    

def test_pure_is_isomorphic_with_relabelled_assemblies(
        assem1: Assembly, component_structures: dict[str, Component]
        ) -> None:
    # Should be isomorphic
    assem2 = deepcopy(assem1)
    assem2.rename_component_ids({'M1': 'M1_'})
    assert pure_is_isomorphic(assem1, assem2, component_structures)


def test_pure_is_isomorphic_with_stereo_isomers(
        assem1: Assembly, assem2: Assembly, 
        component_structures: dict[str, Component]
        ) -> None:
    # Should not be isomorphic, though roughly isomorphic
    assert not pure_is_isomorphic(assem1, assem2, component_structures)



if __name__ == '__main__':
    pytest.main(['-vv', __file__])



import pytest

from nasap_net import Assembly, Component
from nasap_net.reaction_pairing._lib import _are_equivalent_mles
from nasap_net.reaction_pairing.models import _MLE


@pytest.fixture
def components():
    return {
        'L': Component(['a', 'b']),
        'M': Component(['a', 'b']),
        'X': Component(['a']),
    }


def test_inter(components):
    # MX2 + L (-> MLX + X)
    # Although the metal and leaving binding sites are different,
    # the pair of binding sites (metal, leaving) are equivalent, i.e.,
    # (M0.a, X0.a) is equivalent to (M0.b, X1.a).
    MX2 = Assembly(
        {'M0': 'M', 'X0': 'X', 'X1': 'X'},
        [('X0.a', 'M0.a'), ('M0.b', 'X1.a')]
        )
    L = Assembly({'L0': 'L'})
    assert _are_equivalent_mles(
        MX2, L,
        _MLE('M0.a', 'X0.a', 'L0.a'),
        _MLE('M0.b', 'X1.a', 'L0.a'),
        components)


def test_intra(components):
    # Add an additional assembly M2L3
    # M2L3: (a)L0(b)-(a)M0(b)-(a)L1(b)-(a)M1(b)-(a)L2(b)
    M2L3 = Assembly(
        {'M0': 'M', 'M1': 'M', 'L0': 'L', 'L1': 'L', 'L2': 'L'},
        [('L0.b', 'M0.a'), ('M0.b', 'L1.a'), ('L1.b', 'M1.a'),
         ('M1.b', 'L2.a')]
        )

    # M2L3 (-> M2L2-ring + L)
    # Although the metal and leaving binding sites are different,
    # the trio of binding sites (metal, leaving, entering) are equivalent,
    # i.e., (M0.a, L0.b, L2.b) is equivalent to (M1.b, L2.a, L0.a).
    assert _are_equivalent_mles(
        M2L3, None,
        _MLE('M0.a', 'L0.b', 'L2.b'),
        _MLE('M1.b', 'L2.a', 'L0.a'),
        components)



from collections.abc import Iterable, Mapping

from nasap_net import Assembly
from nasap_net.algorithms.subassembly import bond_induced_sub_assembly

__all__ = ['convert_bondset_to_assembly']


def convert_bondset_to_assembly(
        bond_subset: Iterable[str],
        comp_id_to_kind: Mapping[str, str],
        bond_id_to_bindsites: Mapping[str, Iterable[str]]
        ) -> Assembly:
    """Converts the connected bonds to a graph."""
    
    connected_bindsite_pairs = {
        frozenset(bond_id_to_bindsites[bond]) for bond in bond_subset}
    
    template = Assembly(comp_id_to_kind, bond_id_to_bindsites.values())
    
    return bond_induced_sub_assembly(template, connected_bindsite_pairs)



import pytest

from nasap_net.utils.frozen_unordered_pair import FrozenUnorderedPair


def test_init_with_two_elements():
    pair = FrozenUnorderedPair(1, 2)
    assert pair.first == 1
    assert pair.second == 2


def test_init_with_iterable():
    pair = FrozenUnorderedPair([1, 2])
    assert pair.first == 1
    assert pair.second == 2


def test_init_with_invalid_iterable_length():
    with pytest.raises(ValueError):
        FrozenUnorderedPair([1])


def test_init_with_invalid_number_of_arguments():
    with pytest.raises(ValueError):
        FrozenUnorderedPair(1, 2, 3)  # type: ignore


def test_equality():
    pair1 = FrozenUnorderedPair(1, 2)
    pair2 = FrozenUnorderedPair(2, 1)
    assert pair1 == pair2


def test_inequality():
    pair1 = FrozenUnorderedPair(1, 2)
    pair2 = FrozenUnorderedPair(1, 3)
    assert pair1 != pair2


def test_hash():
    pair1 = FrozenUnorderedPair(1, 2)
    pair2 = FrozenUnorderedPair(2, 1)
    assert hash(pair1) == hash(pair2)


def test_repr():
    pair = FrozenUnorderedPair(1, 2)
    assert repr(pair) == 'FrozenUnorderedPair(1, 2)'


def test_iteration():
    pair = FrozenUnorderedPair(1, 2)
    elements = list(pair)
    assert elements == [1, 2]


if __name__ == '__main__':
    pytest.main(['-vv', __file__])



from .chain_map import *
from .circular_perm_comparison import *
from .cyclic_perm import *
from .frozen_unordered_pair import *
from .mapping_iterable_comparison import *
from .nested_dict_update import *
from .node_equivalence import *
from .union_find import UnionFind



from nasap_net import Assembly
from nasap_net.algorithms.assembly_connectivity import extract_connected_assemblies

__all__ = ['separate_product_if_possible']


def separate_product_if_possible(
        assembly: Assembly, component_to_be_in_main: str
        ) -> tuple[Assembly, Assembly | None]:
    main_assembly = None
    leaving_assembly = None

    assems = list(extract_connected_assemblies(assembly))
    assert len(assems) in (1, 2)

    if len(assems) == 1:
        return assembly, None

    # Case where leaving assembly exists
    assem1, assem2 = assems
    if component_to_be_in_main in assem1.component_ids:
        return assem1, assem2
    else:
        assert component_to_be_in_main in assem2.component_ids
        return assem2, assem1



import re

__all__ = ['validate_name_of_component_kind']


def validate_name_of_component_kind(component_type: str) -> None:
    """Check if the component type is valid.

    The component type should be a string that follows the pattern 
    '[A-Za-z_-]+'.
    """
    if not re.fullmatch(r'[A-Za-z_-]+', component_type):
        raise ValueError(
            'The component type should follow the pattern [A-Za-z_-]+.')



import os
from collections.abc import Hashable, Iterable, Mapping, Sequence
from typing import Literal, TypeVar

from nasap_net import enum_bond_subsets
from nasap_net.pipelines.lib import read_file, write_output
from nasap_net.utils import cyclic_perms_to_map, resolve_chain_map

_T = TypeVar('_T', bound=Hashable)



# ============================================================
# Main Process
# ============================================================

def enum_bond_subsets_pipeline(
        input_path: os.PathLike | str,
        output_path: os.PathLike | str,
        *,
        overwrite: bool = False,
        verbose: bool = False,
        path_to_output_resolved_sym_ops: os.PathLike | str | None = None
        ) -> None:
    """Enumerate bond subsets and save the result to a file.

    This function is a pipeline function that reads an input file,
    enumerates connected subsets of bonds excluding symmetry-equivalent
    ones, and saves the result to an output file.

    The core function that enumerates bond subsets is 
    `nasap_net.enum_bond_subsets`.

    See "Notes" section for details on the enumeration logic.

    Parameters
    ----------
    input_path : os.PathLike or str
        Path to the input file.
    output_path : os.PathLike or str
        Path to the output file.
    overwrite : bool, optional
        If True, overwrite the output file if it exists. Default is False.
    verbose : bool, optional
        If True, print detailed execution messages. Default is False.
    resolved_sym_ops : os.PathLike or str or None, optional
        If provided, save resolved symmetry operation mappings to the
        specified file in YAML format. Default is None.

    See Also
    --------
    nasap_net.enum_bond_subsets :
        The core function that enumerates bond subsets. 
        This function performs the actual enumeration logic, 
        while `enum_bond_subsets_pipeline` handles file I/O.

    Notes
    -----
    Functionality:
        When an assembly is represented as a set of bonds, subsets of 
        the bonds can be considered as fragments of the assembly.
        Considering the M2L3 linear assembly, L-M-L-M-L, with bonds named
        1, 2, 3, and 4 from left to right, for example, the assembly is 
        represented as {1, 2, 3, 4}, and its subsets such as {1}, {1, 2}, 
        and {2, 3} can be considered as fragments of the assembly, i.e.,
        L-M, L-M-L, and M-L-M, respectively.
        
        In this function, such subsets of bonds are enumerated
        under the following conditions:

        - Only connected subsets are enumerated.
        In the previous example, {1, 3} is not considered as a fragment.

        - Symmetry-equivalent fragments are excluded.
        If two fragments are superimposable under at least one symmetry 
        operation provided as `sym_ops`, only one of them is included 
        in the result.
        In the previous example, {1, 2} and {3, 4} are symmetry-equivalent
        under the C2 operation, thus only one of them is included in the
        result.

        Note that the result includes the assembly itself as a fragment,
        i.e., {1, 2, 3, 4} in the previous example.

        In summary, this function enumerates *connected* subsets of bonds
        *excluding symmetry-equivalent ones*.
    
    Style of Input File:
        The input file should be in YAML format.
        The required keys are as follows:
        - "bonds": A list of bond IDs.
        - "bond_adjacency": A dictionary mapping a bond to its adjacent bonds.
        - "sym_ops": A dictionary mapping a symmetry operation name to
          a definition. The definition can be a mapping, a list of cyclic
          permutations, or a list of operation names forming a chain.
          Example::

            sym_ops:
                op1: {1: 2, 2: 3, 3: 1}
                op2: [[1, 2], [3]]  # equivalent to {1: 2, 2: 1, 3: 3}
                op3: ['op2', 'op1']  # op3(x) = op2(op1(x)) i.e., {1: 1, 2: 3, 3: 2}
    
    Example of Input File::

        bonds: [1, 2, 3, 4]
        bond_adjacency:
            1: [2]
            2: [1, 3]
            3: [2, 4]
            4: [3]
        sym_ops:
            C2: {1: 4, 2: 3, 3: 2, 4: 1}

    Style of Output File:
        The output file will be a dictionary with integer keys 
        starting from 0. Each key maps to a list of bond IDs representing 
        a connected subset of bonds. The subsets are sorted in a consistent 
        order based on the length of each subset and the elements within 
        each subset.
    
    Example of Output File::
    
        0: [1]
        1: [2]
        2: [1, 2]
        3: [2, 3]
        4: [1, 2, 3]
        5: [1, 2, 3, 4]
    """
    # Input
    input_data = read_file(input_path, verbose=verbose)
    validate_bonds(input_data['bonds'])
    validate_bond_adjacency(input_data['bond_adjacency'], input_data['bonds'])
    if 'sym_ops' in input_data:
        sym_ops = parse_sym_ops(input_data['sym_ops'])
    else:
        sym_ops = None
    if sym_ops is not None:
        validate_sym_ops(sym_ops, input_data['bonds'])
    
    # Save sym_ops to WIP
    if path_to_output_resolved_sym_ops:
        write_output(
            path_to_output_resolved_sym_ops, sym_ops, overwrite=overwrite,
            verbose=verbose, header='Resolved symmetry operations')
    
    # Main process
    if verbose:
        print('Enumerating bond subsets...')
    result = enum_bond_subsets(
        input_data['bonds'],
        input_data['bond_adjacency'],
        sym_ops)
    if verbose:
        print('Finished enumeration.')
        print(f'Number of bond subsets: {len(result)}')
    
    # Output
    formatted_result = sort_bond_subsets(result)

    id_to_bondset = {
        id_: bondset for id_, bondset in enumerate(formatted_result)
        }
    
    write_output(
        output_path, id_to_bondset, overwrite=overwrite, verbose=verbose,
        header='Enumerated bond subsets')


# ============================================================
# Input Processing
# ============================================================

def validate_bonds(bonds: list) -> None:
    if len(bonds) != len(set(bonds)):
        raise ValueError('"bonds" must not contain duplicates.')


def validate_bond_adjacency(bond_to_adj_bonds: Mapping, bonds: list) -> None:
    if set(bond_to_adj_bonds.keys()) != set(bonds):
        raise ValueError('Keys in "bond_adjacency" must be the same as "bonds".')
    
    bonds_set = set(bonds)
    for bond, adj_bonds in bond_to_adj_bonds.items():
        if not all(bond in bonds_set for bond in adj_bonds):
            raise ValueError(
                f'All elements in "bond_adjacency" must be in "bonds". '
                f'Error at bond: {bond}.')


def parse_sym_ops(op_data: Mapping) -> Mapping:
    if not op_data:
        return {}
    
    resolved_ops = {}
    chain_ops = {}
    for op_name, op_def in op_data.items():
        def_type = op_def_type(op_def)
        if def_type == 'map':
            resolved_ops[op_name] = op_def
        elif def_type == 'perms':
            resolved_ops[op_name] = cyclic_perms_to_map(op_def)
        elif def_type == 'chain':
            chain_ops[op_name] = op_def

    # Resolve chain operations
    while chain_ops:
        for op_name, op_def in chain_ops.items():
            try:
                resolved_ops[op_name] = resolve_sym_op_chain(
                    op_def, resolved_ops)
            except ChainOpResolveError:
                continue
            del chain_ops[op_name]
            break
        else:  # No progress
            raise ValueError('Cannot resolve chain operations.')

    assert set(resolved_ops.keys()) == set(op_data.keys())
    assert chain_ops == {}
    return resolved_ops


def op_def_type(
        op_def: Mapping[_T, _T] | list[list[_T]] | list[str]) -> Literal['map', 'perms', 'chain']:
    if isinstance(op_def, Mapping):  # e.g. {1: 1, 2: 3, 3: 2}
        return 'map'
    elif isinstance(op_def, list):  # e.g. [[1], [2, 3]] or ['op2', 'op1']
        if isinstance(op_def[0], list):  # e.g. [[1], [2, 3]]
            return 'perms'
        else:  # e.g. ['op2', 'op1']
            return 'chain'


class ChainOpResolveError(Exception):
    pass


def resolve_sym_op_chain(
        op_chain: list[str], resolved_ops: Mapping[str, Mapping[_T, _T]]
        ) -> Mapping[_T, _T]:
    if set(op_chain) <= set(resolved_ops.keys()):  # Can resolve
        return resolve_chain_map(
            *[resolved_ops[op] for op in reversed(op_chain)])
    raise ChainOpResolveError


def validate_sym_ops(
        sym_ops: Mapping, bonds: Iterable[_T]
        ) -> None:
    bonds = set(bonds)
    for op_name, op_map in sym_ops.items():
        if set(op_map.keys()) != bonds:
            raise ValueError(
                f'Keys in the operation mapping must contain all bonds. '
                f'Error at operation: {op_name}.')
        if set(op_map.values()) != bonds:
            raise ValueError(
                f'Values in the operation mapping must contain all bonds. '
                f'Error at operation: {op_name}.')


# ============================================================
# Result Formatting and Saving
# ============================================================

def sort_bond_subsets(
        bond_subsets: Iterable[Iterable[_T]]) -> list[list[_T]]:
    """Sort bond subsets for deterministic output.

    This function takes bond subsets and returns them sorted 
    in a consistent order based on the length of each subset 
    and the elements within each subset.

    Parameters
    ----------
    bond_subsets : Iterable[Iterable[_T]]
        The input bond subsets to sort.

    Returns
    -------
    list[list[_T]]
        A sorted list of sorted bond subsets.    
    """
    return sorted(
        [sorted(bondset) for bondset in bond_subsets if bondset],
        key=lambda x: (len(x), x))



import pytest
import yaml

import nasap_net as rx
from nasap_net import AuxEdge


def test_LocalAuxEdge_initialization():
    edge = AuxEdge('site1', 'site2', 'cis')
    assert edge.local_binding_site1 == 'site1'
    assert edge.local_binding_site2 == 'site2'
    assert edge.aux_kind == 'cis'


def test_LocalAuxEdge_same_binding_sites():
    with pytest.raises(ValueError):
        AuxEdge('site1', 'site1', 'cis')


def test_LocalAuxEdge_equality():
    edge1 = AuxEdge('site1', 'site2', 'cis')
    edge2 = AuxEdge('site2', 'site1', 'cis')
    assert edge1 == edge2


def test_LocalAuxEdge_inequality():
    edge1 = AuxEdge('site1', 'site2', 'cis')
    edge2 = AuxEdge('site1', 'site3', 'cis')
    assert edge1 != edge2


def test_LocalAuxEdge_hash():
    edge1 = AuxEdge('site1', 'site2', 'cis')
    edge2 = AuxEdge('site2', 'site1', 'cis')
    assert hash(edge1) == hash(edge2)


def test_LocalAuxEdge_repr():
    edge = AuxEdge('site1', 'site2', 'cis')
    assert repr(edge) == "AuxEdge('site1', 'site2', 'cis')"


def test_lt():
    edge1 = AuxEdge('d', 'a', 'cis')
    edge2 = AuxEdge('b', 'c', 'cis')
    assert edge1 < edge2
    assert not edge2 < edge1


def test_yaml():
    edge = AuxEdge('a', 'b', 'cis')
    edge_yaml = yaml.dump(edge, default_flow_style=None)
    loaded_edge = yaml.safe_load(edge_yaml)
    assert edge == loaded_edge


if __name__ == '__main__':
    pytest.main(['-vv', __file__])



from abc import ABC, abstractmethod
from collections.abc import Mapping
from dataclasses import dataclass
from functools import cached_property
from typing import ClassVar

from .assembly import Assembly


class RichReactionBase(ABC):
    @property
    @abstractmethod
    def metal_kind(self) -> str:
        """Kind of the metal component."""
        pass

    @property
    @abstractmethod
    def leaving_kind(self) -> str:
        """Kind of the leaving component."""
        pass

    @property
    @abstractmethod
    def entering_kind(self) -> str:
        """Kind of the entering component."""
        pass

    @classmethod
    @abstractmethod
    def from_reaction(cls, reaction, id_to_assembly: Mapping[int, Assembly]):
        """Create a RichReaction from a Reaction and a mapping from assembly
        IDs to assemblies."""
        pass


@dataclass(frozen=True)
class InterReactionRich(RichReactionBase):
    init_assem: Assembly
    entering_assem: Assembly
    product_assem: Assembly
    leaving_assem: Assembly | None
    metal_bs: str
    leaving_bs: str
    entering_bs: str
    duplicate_count: int

    @cached_property
    def metal_kind(self) -> str:
        return self.init_assem.get_component_kind_of_bindsite(
            self.metal_bs)

    @cached_property
    def leaving_kind(self) -> str:
        return self.init_assem.get_component_kind_of_bindsite(
            self.leaving_bs)

    @cached_property
    def entering_kind(self) -> str:
        return self.entering_assem.get_component_kind_of_bindsite(
            self.entering_bs)

    @classmethod
    def from_reaction(cls, reaction, id_to_assembly: Mapping[int, Assembly]):
        init_assem = id_to_assembly[reaction.init_assem_id]
        entering_assem = id_to_assembly[reaction.entering_assem_id]
        product_assem = id_to_assembly[reaction.product_assem_id]
        if reaction.leaving_assem_id is None:
            leaving_assem = None
        else:
            leaving_assem = id_to_assembly[reaction.leaving_assem_id]
        return cls(
            init_assem=init_assem,
            entering_assem=entering_assem,
            product_assem=product_assem,
            leaving_assem=leaving_assem,
            metal_bs=reaction.metal_bs,
            leaving_bs=reaction.leaving_bs,
            entering_bs=reaction.entering_bs,
            duplicate_count=reaction.duplicate_count
        )


@dataclass(frozen=True)
class IntraReactionRich(RichReactionBase):
    init_assem: Assembly
    entering_assem: ClassVar[None] = None
    product_assem: Assembly
    leaving_assem: Assembly | None
    metal_bs: str
    leaving_bs: str
    entering_bs: str
    duplicate_count: int

    @cached_property
    def metal_kind(self) -> str:
        return self.init_assem.get_component_kind_of_bindsite(
            self.metal_bs)

    @cached_property
    def leaving_kind(self) -> str:
        return self.init_assem.get_component_kind_of_bindsite(
            self.leaving_bs)

    @cached_property
    def entering_kind(self) -> str:
        return self.init_assem.get_component_kind_of_bindsite(
            self.entering_bs)

    @classmethod
    def from_reaction(cls, reaction, id_to_assembly: Mapping[int, Assembly]):
        init_assem = id_to_assembly[reaction.init_assem_id]
        product_assem = id_to_assembly[reaction.product_assem_id]
        if reaction.leaving_assem_id is None:
            leaving_assem = None
        else:
            leaving_assem = id_to_assembly[reaction.leaving_assem_id]
        return cls(
            init_assem=init_assem,
            product_assem=product_assem,
            leaving_assem=leaving_assem,
            metal_bs=reaction.metal_bs,
            leaving_bs=reaction.leaving_bs,
            entering_bs=reaction.entering_bs,
            duplicate_count=reaction.duplicate_count
        )



from itertools import product

import networkx as nx
import pytest

from nasap_net.utils.node_equivalence import group_equivalent_nodes_or_nodesets


@pytest.fixture
def graph_data():
    G = nx.Graph()
    G.add_nodes_from(['1', '2', '3', '4', '5'])
    G.add_edges_from([('1', '2'), ('1', '3'), ('1', '4'), ('1', '5')])
    
    ISOMORPHISMS = [
        {'1': '1', '2': '3', '3': '4', '4': '5', '5': '2'},  # C_4
    ]
    
    return G, ISOMORPHISMS


def test_node_equivalence_single_nodes(graph_data):
    G, ISOMORPHISMS = graph_data
    NODES = {'1', '2', '3', '4', '5'}
    
    node_equiv = group_equivalent_nodes_or_nodesets(NODES, ISOMORPHISMS)
    
    assert node_equiv == {frozenset(['1']), frozenset(['2', '3', '4', '5'])}


def test_node_equivalence_node_sets(graph_data):
    G, ISOMORPHISMS = graph_data
    NODESETS = [('1', '2'), ('2', '3'), ('1', '3'), ('1', '1'), ('2', '2')]

    nodeset_equiv = group_equivalent_nodes_or_nodesets(NODESETS, ISOMORPHISMS)

    assert nodeset_equiv == {
        frozenset([('1', '2'), ('1', '3')]),
        frozenset([('2', '3')]),
        frozenset([('1', '1')]),
        frozenset([('2', '2')]),
    }


def test_node_equivalence_product_node_sets(graph_data):
    G, ISOMORPHISMS = graph_data
    NODES = {'1', '2', '3', '4', '5'}
    NODESETS2 = product(NODES, repeat=2)  # total 25 pairs

    nodeset_equiv2 = group_equivalent_nodes_or_nodesets(NODESETS2, ISOMORPHISMS)

    assert nodeset_equiv2 == {
        frozenset([('1', '1')]),
        frozenset([('1', '2'), ('1', '3'), ('1', '4'), ('1', '5')]),
        frozenset([('2', '1'), ('3', '1'), ('4', '1'), ('5', '1')]),
        frozenset([('2', '2'), ('3', '3'), ('4', '4'), ('5', '5')]),
        frozenset([('2', '3'), ('3', '4'), ('4', '5'), ('5', '2')]),
        frozenset([('2', '4'), ('3', '5'), ('4', '2'), ('5', '3')]),
        frozenset([('2', '5'), ('3', '2'), ('4', '3'), ('5', '4')]),
    }


if __name__ == '__main__':
    pytest.main(['-v', __file__])



from typing import TypeAlias

import pandas as pd
import pytest
import yaml

from nasap_net import Assembly, Component, InterReaction, IntraReaction
from nasap_net.pipelines import explore_reactions_pipeline

Reaction: TypeAlias = InterReaction | IntraReaction

def test_comprehensive_reaction_sets(tmpdir):
    components_data = {
        'component_kinds': {
        'M': Component(['a', 'b']),
        'L': Component(['a', 'b']),
        'X': Component(['a']),
        }
    }

    id_to_assembly = {
        # MX2: X0(a)-(a)M0(b)-(a)X1
        0: Assembly(
            {'M0': 'M', 'X0': 'X', 'X1': 'X'},
            [('X0.a', 'M0.a'), ('M0.b', 'X1.a')]
        ),
        1: Assembly({'L0': 'L'}),  # L: (a)L0(b)
        2: Assembly({'X0': 'X'}),  # X: (a)X0
        # MLX: (a)L0(b)-(a)M0(b)-(a)X0
        3: Assembly(
            {'M0': 'M', 'L0': 'L', 'X0': 'X'},
            [('L0.b', 'M0.a'), ('M0.b', 'X0.a')]
        ),
        # ML2: (a)L0(b)-(a)M0(b)-(a)L1(b)
        4: Assembly(
            {'M0': 'M', 'L0': 'L', 'L1': 'L'},
            [('L0.b', 'M0.a'), ('M0.b', 'L1.a')]
        ),
        # M2L2X: X0(a)-(a)M0(b)-(a)L0(b)-(a)M1(b)-(a)L1(b)
        5: Assembly(
            {'M0': 'M', 'M1': 'M', 'L0': 'L', 'L1': 'L', 'X0': 'X'},
            [('X0.a', 'M0.a'), ('M0.b', 'L0.a'), ('L0.b', 'M1.a'),
             ('M1.b', 'L1.a')]
        ),
        # M2LX2: X0(a)-(a)M0(b)-(a)L0(b)-(a)M1(b)-(a)X1
        6: Assembly(
            {'M0': 'M', 'M1': 'M', 'L0': 'L', 'X0': 'X', 'X1': 'X'},
            [('X0.a', 'M0.a'), ('M0.b', 'L0.a'), ('L0.b', 'M1.a'),
             ('M1.b', 'X1.a')]
        ),
        # M2L2-ring: //-(a)M0(b)-(a)L0(b)-(a)M1(b)-(a)L1(b)-//
        7: Assembly(
            {'M0': 'M', 'M1': 'M', 'L0': 'L', 'L1': 'L'},
            [('M0.b', 'L0.a'), ('L0.b', 'M1.a'), ('M1.b', 'L1.a'),
             ('L1.b', 'M0.a')]
        ),
    }

    config = {
        'mle_kinds': {
            'metal': 'M',
            'leaving': 'X',
            'entering': 'L',
        }
    }

    # Create input files
    assemblies_path = tmpdir / 'assemblies.yaml'
    components_path = tmpdir / 'components.yaml'
    config_path = tmpdir / 'config.yaml'
    output_path = tmpdir / 'reactions.csv'

    with open(assemblies_path, 'w') as f:
        yaml.dump(id_to_assembly, f)
    with open(components_path, 'w') as f:
        yaml.dump(components_data, f)
    with open(config_path, 'w') as f:
        yaml.dump(config, f)

    expected_csv = """\
init_assem_id,entering_assem_id,product_assem_id,leaving_assem_id,metal_bs,leaving_bs,entering_bs,duplicate_count
5,,7,2,M0.a,X0.a,L1.b,1
0,1,3,2,M0.a,X0.a,L0.a,4
0,3,6,2,M0.a,X0.a,L0.a,2
0,4,5,2,M0.a,X0.a,L0.a,4
3,1,4,2,M0.b,X0.a,L0.a,2
3,3,5,2,M0.b,X0.a,L0.a,2
6,1,5,2,M0.a,X0.a,L0.a,4
"""

    # Run the pipeline
    explore_reactions_pipeline(
        assemblies_path=assemblies_path,
        components_path=components_path,
        config_path=config_path,
        output_path=output_path,
        overwrite=True, verbose=True
    )

    # Read the output CSV
    with open(output_path, 'r') as f:
        output_csv = f.read()

    # Check if the output CSV matches the expected CSV
    assert output_csv == expected_csv


if __name__ == "__main__":
    pytest.main([__file__, '-v'])



from collections.abc import Iterable, Sequence


# TODO: Add validation
# TODO: Use generic types
def cyclic_perm_to_map(
        cyclic_permutation: Sequence[str]) -> dict[str, str]:
    mapping: dict[str, str] = {}
    for i, source in enumerate(cyclic_permutation):
        # (i + 1) % N maps as follows:
        # 0 -> 1, 1 -> 2, ..., N-1 -> 0
        target = cyclic_permutation[(i + 1) % len(cyclic_permutation)]
        mapping[source] = target
    return mapping


def cyclic_perms_to_map(
        cyclic_permutations: Iterable[Sequence[str]]) -> dict[str, str]:
    mapping = {}
    for perm in cyclic_permutations:
        mapping.update(cyclic_perm_to_map(perm))
    return mapping



import pytest

from nasap_net import Assembly, InterReactionRich, IntraReactionRich
from nasap_net.reaction_classification.utils.inter_or_intra_judgement import \
    inter_or_intra


def test_with_intra_reaction():
    reaction = IntraReactionRich(
        Assembly(), Assembly(), Assembly(), 
        '', '', '', 1
    )
    result = inter_or_intra(reaction)
    assert result == "intra"


def test_with_inter_reaction():
    reaction = InterReactionRich(
        Assembly(), Assembly(), Assembly(), Assembly(),
        '', '', '', 1
    )
    result = inter_or_intra(reaction)
    assert result == "inter"


if __name__ == '__main__':
    pytest.main(['-vv', __file__])



import pytest

from nasap_net.classes.validations import validate_name_of_component_kind


def test_validate_name_of_component_kind():
    pass


if __name__ == '__main__':
    pytest.main(['-vv', __file__])



import os
from typing import Any

import yaml


def write_output(
        output_path: os.PathLike | str, data: Any,
        *,
        overwrite: bool = False,
        default_flow_style: bool | None = None,
        verbose: bool = False,
        header: str | None = None,
        ) -> None:
    """Write the output data to a file using YAML format."""
    if not overwrite and os.path.exists(output_path):
        raise FileExistsError(f'Output file "{output_path}" already exists.')
    if verbose:
        print(f'Saving the results to "{output_path}"...')
    dir_ = os.path.dirname(output_path)
    if dir_:
        os.makedirs(os.path.dirname(output_path), exist_ok=True)
    with open(output_path, 'w') as f:
        yaml.dump(data, f, default_flow_style=default_flow_style)
    if verbose:
        if header:
            print(f'{header}: Successfully saved to "{output_path}".')
        else:
            print(f'Successfully saved to "{output_path}".')



from collections.abc import Hashable, Mapping
from dataclasses import dataclass

import igraph as ig
from bidict import frozenbidict
from frozendict import frozendict

from nasap_net.models import Assembly, BindingSite
from nasap_net.types import ID


@dataclass(frozen=True, init=False)
class GraphConversionResult:
    graph: ig.Graph
    core_mapping: frozenbidict[ID, int]
    binding_site_mapping: frozenbidict[BindingSite, int]

    def __init__(
            self,
            graph: ig.Graph,
            core_mapping: Mapping[ID, int],
            binding_site_mapping: Mapping[BindingSite, int],
    ) -> None:
        object.__setattr__(self, 'graph', graph)
        object.__setattr__(
            self, 'core_mapping', frozenbidict(core_mapping))
        object.__setattr__(
            self, 'binding_site_mapping',
            frozenbidict(binding_site_mapping))


def convert_assembly_to_igraph(assembly: Assembly) -> GraphConversionResult:
    g = ig.Graph()
    core_mapping = {}
    binding_site_mapping = {}

    for comp_id, comp in assembly.components.items():
        # Add the core node
        g.add_vertices(  # More efficient than add_vertex
            1,
            {
                'comp_id': [comp_id],
                'comp_kind': [comp.kind],
                'core_or_site': ['core'],
            }
        )
        core_mapping[comp_id] = g.vcount() - 1

        # Add the binding sites
        g.add_vertices(
            len(comp.site_ids),
            {
                'comp_id': [comp_id] * len(comp.site_ids),
                'comp_kind': [comp.kind] * len(comp.site_ids),
                'core_or_site': ['site'] * len(comp.site_ids),
                'site_id': list(comp.site_ids),
            }
        )
        start_id = g.vcount() - len(comp.site_ids)
        for i, site in enumerate(comp.get_binding_sites(comp_id)):
            binding_site_mapping[site] = start_id + i

        # Add the edges between core and sites
        g.add_edges(
            [(core_mapping[comp_id], binding_site_mapping[site])
                for site in comp.get_binding_sites(comp_id)]
        )

        # Add the auxiliary edges
        for aux in comp.aux_edges:
            site1, site2 = aux.get_binding_sites(comp_id)
            if aux.kind is None:
                g.add_edges(
                    [(
                        binding_site_mapping[site1],
                        binding_site_mapping[site2]
                    )]
                )
            else:
                g.add_edges(
                    [(
                        binding_site_mapping[site1],
                        binding_site_mapping[site2]
                    )],
                    {'aux_kind': [aux.kind]}
                )

    # Add the bonds
    for bond in assembly.bonds:
        site1, site2 = bond.sites
        g.add_edge(
            binding_site_mapping[site1],
            binding_site_mapping[site2]
        )

    return GraphConversionResult(
        graph=g,
        core_mapping=core_mapping,
        binding_site_mapping=binding_site_mapping
    )


def is_isomorphic(assem1: Assembly, assem2: Assembly) -> bool:
    g1 = convert_assembly_to_igraph(assem1).graph
    g2 = convert_assembly_to_igraph(assem2).graph

    try:
        colors = _color_vertices_and_edges(g1, g2)
    except NoIsomorphismFoundError:
        return False

    return g1.isomorphic_vf2(
        g2,
        color1=colors.v_color1,
        color2=colors.v_color2,
        edge_color1=colors.e_color1,
        edge_color2=colors.e_color2,
    )


class NoIsomorphismFoundError(Exception):
    pass


@dataclass(frozen=True, init=False)
class Isomorphism:
    comp_id_mapping: frozendict[ID, ID]
    binding_site_mapping: frozendict[BindingSite, BindingSite]

    def __init__(
            self,
            comp_id_mapping: Mapping[ID, ID],
            binding_site_mapping: Mapping[BindingSite, BindingSite]
    ) -> None:
        object.__setattr__(
            self, 'comp_id_mapping', frozendict(comp_id_mapping))
        object.__setattr__(
            self, 'binding_site_mapping', frozendict(binding_site_mapping))


def get_isomorphism(assem1: Assembly, assem2: Assembly) -> Isomorphism:
    conv_res1 = convert_assembly_to_igraph(assem1)
    conv_res2 = convert_assembly_to_igraph(assem2)

    g1 = conv_res1.graph
    g2 = conv_res2.graph

    try:
        colors = _color_vertices_and_edges(g1, g2)
    except NoIsomorphismFoundError:
        raise NoIsomorphismFoundError() from None

    mapping: list[int]
    _, mapping, _ = g1.isomorphic_vf2(
        g2,
        color1=colors.v_color1,
        color2=colors.v_color2,
        edge_color1=colors.e_color1,
        edge_color2=colors.e_color2,
        return_mapping_12=True,
    )

    return _decode_mapping(mapping, conv_res1, conv_res2)


def get_all_isomorphisms(
        assem1: Assembly, assem2: Assembly) -> set[Isomorphism]:
    conv_res1 = convert_assembly_to_igraph(assem1)
    conv_res2 = convert_assembly_to_igraph(assem2)

    try:
        colors = _color_vertices_and_edges(conv_res1.graph, conv_res2.graph)
    except NoIsomorphismFoundError:
        raise NoIsomorphismFoundError() from None

    res: list[list[int]] = conv_res1.graph.get_isomorphisms_vf2(
        conv_res2.graph,
        color1=colors.v_color1,
        color2=colors.v_color2,
        edge_color1=colors.e_color1,
        edge_color2=colors.e_color2,
    )

    return {_decode_mapping(mapping, conv_res1, conv_res2) for mapping in res}


def _decode_mapping(
        mapping: list[int],
        conv_res1: GraphConversionResult,
        conv_res2: GraphConversionResult,
) -> Isomorphism:
    comp_id_mapping = {}
    binding_site_mapping = {}
    for v1, v2 in enumerate(mapping):
        if v1 in conv_res1.core_mapping.inv:
            comp_id1 = conv_res1.core_mapping.inv[v1]
            comp_id2 = conv_res2.core_mapping.inv[v2]
            comp_id_mapping[comp_id1] = comp_id2
        else:
            assert v1 in conv_res1.binding_site_mapping.inv
            site1 = conv_res1.binding_site_mapping.inv[v1]
            site2 = conv_res2.binding_site_mapping.inv[v2]
            binding_site_mapping[site1] = site2
    return Isomorphism(
        comp_id_mapping=comp_id_mapping,
        binding_site_mapping=binding_site_mapping
    )


@dataclass(frozen=True)
class _Colors:
    v_color1: tuple[int, ...]
    v_color2: tuple[int, ...]
    e_color1: tuple[int, ...]
    e_color2: tuple[int, ...]


def _color_vertices_and_edges(g1: ig.Graph, g2: ig.Graph) -> _Colors:
    try:
        v_color1, v_color2 = _vertex_color_lists(g1, g2)
    except _NotIsomorphicError:
        raise NoIsomorphismFoundError() from None

    try:
        e_color1, e_color2 = _edge_color_lists(g1, g2)
    except _NotIsomorphicError:
        raise NoIsomorphismFoundError() from None

    return _Colors(
        v_color1=tuple(v_color1),
        v_color2=tuple(v_color2),
        e_color1=tuple(e_color1),
        e_color2=tuple(e_color2),
    )


class _NotIsomorphicError(Exception):
    pass


def _vertex_color_lists(
        g1: ig.Graph, g2: ig.Graph) -> tuple[list[int], list[int]]:
    def _readable_v_color(vertex: ig.Vertex) -> Hashable:
        return vertex['core_or_site'], vertex['comp_kind']

    colors1 = {_readable_v_color(v) for v in g1.vs}
    colors2 = {_readable_v_color(v) for v in g2.vs}
    if colors1 != colors2:
        raise _NotIsomorphicError()

    color_to_int: dict[Hashable, int] = {c: i for i, c in enumerate(colors1)}
    color_list1 = [color_to_int[_readable_v_color(v)] for v in g1.vs]
    color_list2 = [color_to_int[_readable_v_color(v)] for v in g2.vs]
    return color_list1, color_list2


def _edge_color_lists(
        g1: ig.Graph, g2: ig.Graph) -> tuple[list[int], list[int]]:
    def _readable_edge_color(edge: ig.Edge) -> Hashable | None:
        return edge['aux_kind'] if 'aux_kind' in edge.attributes() else None

    colors1 = {_readable_edge_color(e) for e in g1.es}
    colors2 = {_readable_edge_color(e) for e in g2.es}
    if colors1 != colors2:
        raise _NotIsomorphicError()

    color_to_int: dict[Hashable | None, int] = {
        c: i for i, c in enumerate(colors1)}
    color_list1 = [color_to_int[_readable_edge_color(e)] for e in g1.es]
    color_list2 = [color_to_int[_readable_edge_color(e)] for e in g2.es]
    return color_list1, color_list2



import os

import pytest
import yaml

from nasap_net.pipelines.bondset_enumeration import enum_bond_subsets_pipeline


def test_without_sym_ops(tmp_path):
    # M2L3 linear: L-M-L-M-L
    # bonds: 1, 2, 3, 4 from left to right
    INPUT_DATA = {
        'bonds': [1, 2, 3, 4],
        'bond_adjacency': {
            1: {2},
            2: {1, 3},
            3: {2, 4},
            4: {3},
        }
    }
    EXPECTED = {
        0: [1], 1: [2], 2: [3], 3: [4],
        4: [1, 2], 5: [2, 3], 6: [3, 4],
        7: [1, 2, 3], 8: [2, 3, 4],
        9: [1, 2, 3, 4]
    }
    input_path = tmp_path / "input.yaml"
    with open(input_path, 'w') as f:
        yaml.safe_dump(INPUT_DATA, f)

    output_path = tmp_path / "output.yaml"
    enum_bond_subsets_pipeline(input_path, output_path)

    with open(output_path, 'r') as f:
        actual_output = yaml.safe_load(f)

    assert actual_output == EXPECTED


def test_with_sym_ops(tmp_path):
    # M2L3 linear: L-M-L-M-L
    # bonds: 1, 2, 3, 4 from left to right
    INPUT_DATA = {
        'bonds': [1, 2, 3, 4],
        'bond_adjacency': {
            1: {2},
            2: {1, 3},
            3: {2, 4},
            4: {3},
        },
        'sym_ops': {
            'C2': {1: 4, 2: 3, 3: 2, 4: 1}
        }
    }
    EXPECTED = {
        0: [1], 1: [2], 2: [1, 2], 3: [2, 3], 4: [1, 2, 3], 5: [1, 2, 3, 4]
    }
    input_path = tmp_path / "input.yaml"
    with open(input_path, 'w') as f:
        yaml.safe_dump(INPUT_DATA, f)
    
    output_path = tmp_path / "output.yaml"
    enum_bond_subsets_pipeline(input_path, output_path)

    with open(output_path, 'r') as f:
        actual_output = yaml.safe_load(f)

    assert actual_output == EXPECTED


def test_missing_input_key(tmp_path):
    INPUT_DATA = {
        'bond_adjacency': {
            1: {2},
            2: {1, 3},
            3: {2, 4},
            4: {3},
        }
    }
    input_path = tmp_path / "input.yaml"
    with open(input_path, 'w') as f:
        yaml.safe_dump(INPUT_DATA, f)

    with pytest.raises(KeyError):
        enum_bond_subsets_pipeline(input_path, "output.yaml")


def test_verbose_false(tmp_path, capsys):
    INPUT_DATA = {
        'bonds': [1, 2, 3, 4],
        'bond_adjacency': {
            1: {2},
            2: {1, 3},
            3: {2, 4},
            4: {3},
        }
    }
    input_path = tmp_path / "input.yaml"
    with open(input_path, 'w') as f:
        yaml.safe_dump(INPUT_DATA, f)

    output_path = tmp_path / "output.yaml"
    enum_bond_subsets_pipeline(input_path, output_path, verbose=False)

    captured = capsys.readouterr()
    assert captured.out == ""
    assert captured.err == ""


def test_verbose_true(tmp_path, capsys):
    INPUT_DATA = {
        'bonds': [1, 2, 3, 4],
        'bond_adjacency': {
            1: {2},
            2: {1, 3},
            3: {2, 4},
            4: {3},
        }
    }
    input_path = tmp_path / "input.yaml"
    with open(input_path, 'w') as f:
        yaml.safe_dump(INPUT_DATA, f)

    output_path = tmp_path / "output.yaml"
    enum_bond_subsets_pipeline(input_path, output_path, verbose=True)

    captured = capsys.readouterr()
    assert f'Reading from "{input_path}"...' in captured.out
    assert f'Finished reading from "{input_path}".' in captured.out
    assert 'Enumerating bond subsets...' in captured.out
    assert f'Number of bond subsets: 10' in captured.out
    assert f'Saving the results to "{output_path}"...' in captured.out
    assert (
        f'Successfully saved to "{output_path}"' in captured.out)


def test_wip(tmp_path):
    INPUT_DATA = {
        'bonds': [1, 2, 3, 4],
        'bond_adjacency': {
            1: {2},
            2: {1, 3},
            3: {2, 4},
            4: {3},
        },
        'sym_ops': {
            'C2': [[1, 4], [2, 3]]
        }
    }
    input_path = tmp_path / "input.yaml"
    with open(input_path, 'w') as f:
        yaml.safe_dump(INPUT_DATA, f)

    output_path = tmp_path / "output.yaml"
    wip_dir = tmp_path / "resolved_sym_ops.yaml"
    enum_bond_subsets_pipeline(input_path, output_path, path_to_output_resolved_sym_ops=wip_dir)

    assert os.path.exists(wip_dir)
    assert os.path.isfile(wip_dir)

    EXPECTED = {
        'C2': {1: 4, 2: 3, 3: 2, 4: 1}
    }
    with open(wip_dir, 'r') as f:
        actual_output = yaml.safe_load(f)
    
    assert actual_output == EXPECTED


if __name__ == "__main__":
    pytest.main(['-v', __file__])



import pytest

from nasap_net.models import Bond, LightAssembly
from nasap_net.yaml.lib import dump_light_assemblies, load_light_assemblies


@pytest.fixture
def MX2():
    return LightAssembly(
        id_='MX2',
        components={'X0': 'X', 'M0': 'M', 'X1': 'X'},
        bonds=[Bond('X0', 0, 'M0', 0), Bond('M0', 1, 'X1', 0)])

@pytest.fixture
def dumped_MX2():
    return """!LightAssembly
components: {M0: M, X0: X, X1: X}
bonds:
- [M0, 0, X0, 0]
- [M0, 1, X1, 0]
id: MX2
"""


def test_dump(MX2, dumped_MX2):
    dumped = dump_light_assemblies(MX2)
    assert dumped == dumped_MX2


def test_load(MX2, dumped_MX2):
    loaded = load_light_assemblies(dumped_MX2)  # type: ignore
    assert loaded == MX2


def test_round_trip():
    assemblies = {
        # MX2: X0(0)-(0)M0(1)-(0)X1
        'MX2': LightAssembly(
            id_='MX2',
            components={'X0': 'X', 'M0': 'M', 'X1': 'X'},
            bonds=[Bond('X0', 0, 'M0', 0), Bond('M0', 1, 'X1', 0)]),
        'free_X': LightAssembly(components={'X0': 'X'}, bonds=[]),
        # MLX: (0)L0(1)-(0)M0(1)-(0)X0
        'MLX': LightAssembly(
            components={'L0': 'L', 'M0': 'M', 'X0': 'X'},
            bonds=[Bond('L0', 1, 'M0', 0), Bond('M0', 1, 'X0', 0)]),
        'M(sq)X4': LightAssembly(
            components={'M0': 'M_square', 'X0': 'X', 'X1': 'X', 'X2': 'X', 'X3': 'X'},
            bonds=[Bond('M0', 0, 'X0', 0), Bond('M0', 1, 'X1', 0),
                   Bond('M0', 2, 'X2', 0), Bond('M0', 3, 'X3', 0)]),
        'M(sq)L2X2': LightAssembly(
            components={'M0': 'M_square', 'L0': 'L', 'L1': 'L', 'X0': 'X', 'X1': 'X'},
            bonds=[Bond('M0', 0, 'X0', 0), Bond('M0', 1, 'L0', 0),
                   Bond('M0', 2, 'X1', 0), Bond('M0', 3, 'L1', 0)]),
    }

    dumped_assemblies = dump_light_assemblies(assemblies)
    loaded_assemblies = load_light_assemblies(dumped_assemblies)
    assert loaded_assemblies == assemblies



from collections.abc import Mapping

from nasap_net import Assembly, Component, is_isomorphic

__all__ = ['is_new']


def is_new(
        hash_: str,
        assembly: Assembly,
        hash_to_uniques: Mapping[str, list[Assembly]],
        component_structures: Mapping[str, Component]
        ) -> bool:
    """Check if the assembly is new."""
    if hash_ not in hash_to_uniques:
        return True
    for unique in hash_to_uniques[hash_]:
        if is_isomorphic(
                assembly, unique, component_structures):
            return False
    return True



from collections.abc import Mapping
from itertools import count
from typing import Literal, overload

from nasap_net import Assembly, Component

from .single_bindsite import cap_single_bindsite

__all__ = ['cap_bindsites']


@overload
def cap_bindsites(
        assembly: Assembly, 
        component_structures: Mapping[str, Component],
        component_kind_to_be_capped: str,
        cap_component_kind: str, cap_bindsite: str,
        copy: Literal[True] = True
        ) -> Assembly: ...
@overload
def cap_bindsites(
        assembly: Assembly, 
        component_structures: Mapping[str, Component],
        component_kind_to_be_capped: str,
        cap_component_kind: str, cap_bindsite: str,
        copy: Literal[False]
        ) -> None: ...
def cap_bindsites(
        assembly: Assembly, 
        component_structures: Mapping[str, Component],
        component_kind_to_be_capped: str,
        cap_component_kind: str, cap_bindsite: str,
        copy: bool = True
        ) -> Assembly | None:
    """Add leaving ligands to all free bindsites of a specific component kind."""
    if copy:
        assembly = assembly.deepcopy()

    target_bindsites: list[str] = []
    for comp_id, comp_kind in assembly.comp_id_to_kind.items():
        if comp_kind != component_kind_to_be_capped:
            continue
        for bindsite in assembly.get_bindsites_of_component(
                comp_id, component_structures):
            if assembly.is_free_bindsite(bindsite):
                target_bindsites.append(bindsite)

    cap_ids = (f'{cap_component_kind}{i}' for i in count())

    for target_bindsite in target_bindsites:
        cap_id = next(cap_ids)
        while cap_id in assembly.component_ids:
            cap_id = next(cap_ids)
        cap_single_bindsite(
            assembly, target_bindsite, cap_id, 
            cap_component_kind, cap_bindsite, copy=False)
    
    if copy:
        return assembly
    else:
        return None



import pytest

from nasap_net import Assembly, Component
from nasap_net.algorithms import are_equivalent_binding_site_lists


@pytest.fixture
def component_structures():
    """Fixture for component structures."""
    return {
        'M_bi': Component(['a', 'b']),
        'M_tetra': Component(['a', 'b', 'c', 'd']),
        'L': Component(['a', 'b']),
        'X': Component(['a']),
    }

@pytest.fixture
def M():
    # (a)M0(b)
    return Assembly({'M0': 'M_bi'})

@pytest.fixture
def ML2():
    # (a)L0(b)-(a)M0(b)-(a)L1(b)
    return Assembly(
        {'M0': 'M_bi', 'L0': 'L', 'L1': 'L'},
        [('L0.b', 'M0.a'), ('M0.b', 'L1.a')]
    )

@pytest.fixture
def MLX():
    # (a)L0(b)-(a)M0(b)-(a)X0
    return Assembly(
        {'M0': 'M_bi', 'L0': 'L', 'X0': 'X'},
        [('L0.b', 'M0.a'), ('M0.b', 'X0.a')]
    )

@pytest.fixture
def ML2X2():
    #              X0
    #             (a)
    #              |
    #             (c)
    # (a)L0(b)--(a)M0(b)--(a)L1(b)
    #             (d)
    #              |
    #             (a)
    #              X1

    # WITHOUT auxiliary edges

    return Assembly(
        {'M0': 'M_tetra', 'L0': 'L', 'L1': 'L', 'X0': 'X', 'X1': 'X'},
        [('M0.a', 'L0.b'), ('M0.b', 'L1.b'),
         ('M0.c', 'X0.a'), ('M0.d', 'X1.a')]
    )


def test_different_lengths_error(M):
    # Check if ValueError is raised for different lengths
    with pytest.raises(ValueError):
        are_equivalent_binding_site_lists(M, ['M0.a'], ['M0.a', 'M0.b'], {})


def test_empty_lists_error(M):
    # Check if ValueError is raised for empty lists
    with pytest.raises(ValueError):
        are_equivalent_binding_site_lists(M, [], [], {})


def test_nonexistent_binding_site_error(M, component_structures):
    # Check if ValueError is raised for nonexistent binding sites
    with pytest.raises(ValueError):
        are_equivalent_binding_site_lists(
            M, ['M0.a'], ['M0.unknown'], component_structures)
    with pytest.raises(ValueError):
        are_equivalent_binding_site_lists(
            M, ['M0.unknown'], ['M0.a'], component_structures)


def test_single_binding_site(M, component_structures):
    assert are_equivalent_binding_site_lists(
        M, ['M0.a'], ['M0.b'], component_structures)


def test_assembly_composed_of_multiple_components(
        ML2, ML2X2, component_structures):
    assert are_equivalent_binding_site_lists(
        ML2, ['L0.a'], ['L1.b'], component_structures)
    assert are_equivalent_binding_site_lists(
        ML2X2, ['L0.a', 'M0.c'], ['L1.a', 'M0.d'], component_structures)


def test_consideration_of_component_kinds(MLX, component_structures):
    # Should not be equivalent
    # because isomorphisms which maps L0 to X0 are not allowed.
    assert not are_equivalent_binding_site_lists(
        MLX, ['L0.b', 'M0.a'], ['X0.a', 'M0.b'], component_structures)


def test_self_comparison(M, ML2, MLX, component_structures):
    assert are_equivalent_binding_site_lists(
        M, ['M0.a'], ['M0.a'], component_structures)
    assert are_equivalent_binding_site_lists(
        ML2, ['L0.a', 'L0.b'], ['L0.a', 'L0.b'], component_structures)
    assert are_equivalent_binding_site_lists(
        MLX, ['L0.a', 'L0.b'], ['L0.a', 'L0.b'], component_structures)


def test_pair(ML2, component_structures):
    # Same binding sites in M
    assert are_equivalent_binding_site_lists(
        ML2, ['L0.a', 'L0.b'], ['L1.b', 'L1.a'], component_structures)
    assert not are_equivalent_binding_site_lists(
        ML2, ['L0.a', 'L0.b'], ['L1.a', 'L1.b'], component_structures)


def test_trio(ML2X2, component_structures):
    # Consider inter-molecular reactions of ML2X2 as an example.
    # trio: (metal, leaving, entering)
    assert are_equivalent_binding_site_lists(
        ML2X2, ['M0.c', 'X0.a', 'L0.a'], ['M0.d', 'X1.a', 'L1.a'],
        component_structures)


def test_auxiliary_edges():
    COMPONENT_KINDS = {
        'M_plain': Component(['a', 'b', 'c', 'd']),
        'M_aux': Component(
            ['a', 'b', 'c', 'd'],
            [('a', 'b', 'cis'), ('b', 'c', 'cis'), 
             ('c', 'd', 'cis'), ('d', 'a', 'cis')]
        ),
        'L': Component(['a', 'b']),
        'X': Component(['a']),
    }

    #           X0
    #          (a)
    #           |
    #          (a)
    # X1(a)--(b)M0(d)--(a)L0(b)
    #          (c)
    #           |
    #          (a)
    #           X2
    MLX3_plain = Assembly(
        {'M0': 'M_plain', 'L0': 'L', 'X0': 'X', 'X1': 'X', 'X2': 'X'},
        [('M0.a', 'X0.a'), ('M0.b', 'X1.a'),
         ('M0.c', 'X2.a'), ('M0.d', 'L0.a')]
        )
    MLX3_aux = Assembly(
        {'M0': 'M_aux', 'L0': 'L', 'X0': 'X', 'X1': 'X', 'X2': 'X'},
        [('M0.a', 'X0.a'), ('M0.b', 'X1.a'),
         ('M0.c', 'X2.a'), ('M0.d', 'L0.a')]
        )
    
    assert are_equivalent_binding_site_lists(
        MLX3_plain, ['M0.d', 'M0.a'], ['M0.d', 'M0.b'], COMPONENT_KINDS)
    assert not are_equivalent_binding_site_lists(
        MLX3_aux, ['M0.d', 'M0.a'], ['M0.d', 'M0.b'], COMPONENT_KINDS)
    
    # Intra-molecular reaction
    assert are_equivalent_binding_site_lists(
        MLX3_plain, ['M0.a', 'X0.a', 'L0.b'], ['M0.b', 'X1.a', 'L0.b'],
        COMPONENT_KINDS)
    assert not are_equivalent_binding_site_lists(
        MLX3_aux, ['M0.a', 'X0.a', 'L0.b'], ['M0.b', 'X1.a', 'L0.b'],
        COMPONENT_KINDS)


if __name__ == "__main__":
    pytest.main([__file__, '-v'])



import pytest

from nasap_net.classes.bindsite_id_converter import BindsiteIdConverter


def test_local_to_global():
    id_converter = BindsiteIdConverter()
    result = id_converter.local_to_global('M1', 'a')
    assert result == 'M1.a'


def test_global_to_local():
    id_converter = BindsiteIdConverter()
    result = id_converter.global_to_local('M1.a')
    assert result == ('M1', 'a')


def test_global_to_local_invalid_format():
    id_converter = BindsiteIdConverter()
    with pytest.raises(AssertionError):
        id_converter.global_to_local('invalid_format')


if __name__ == '__main__':
    pytest.main(['-vv', __file__])



from collections.abc import Mapping
from itertools import cycle

from nasap_net import Assembly

__all__ = ['make_complete_color_map']


def make_complete_color_map(
        assembly: Assembly, color_map: Mapping[str, str]
        ) -> dict[str, str]:
    color_undefined_component_kinds = (
        assembly.component_kinds - set(color_map.keys()))
    if not color_undefined_component_kinds:
        return dict(color_map)

    # infinity iterator
    DEFAULT_COLORS = cycle([
        '#FE7F6E', '#C89F38', '#4CBB51', '#53AFDC', '#9A9EEC',
        '#9AA6B5', '#A5A5A5'])

    # Convert to dict to allow modification
    complete_map = dict(color_map)

    # Use sorted to make the order of colors deterministic
    for comp_kind in sorted(color_undefined_component_kinds):
        complete_map[comp_kind] = next(DEFAULT_COLORS)

    return complete_map


import pytest

from nasap_net import Assembly, Component
from nasap_net.duplicate_exclusion import group_assemblies_by_isomorphism


def test_group_assemblies_by_isomorphism():
    id_to_graph = {
        0: Assembly(  # MLX-1
            {'M1': 'M', 'L1': 'L', 'X1': 'X'}, 
            [('M1.a', 'L1.a'), ('M1.b', 'X1.a')]),
        1: Assembly(  # MLX-2
            {'M2': 'M', 'L2': 'L', 'X2': 'X'}, 
            [('M2.a', 'L2.a'), ('M2.b', 'X2.a')]),
        2: Assembly(  # MLX-3
            {'M3': 'M', 'L3': 'L', 'X3': 'X'}, 
            [('M3.a', 'L3.a'), ('M3.b', 'X3.a')]),
        3: Assembly(  # MX2
            {'M1': 'M', 'X1': 'X', 'X2': 'X'}, 
            [('M1.a', 'X1.a'), ('M1.b', 'X2.a')]),
        4: Assembly(  # L
            {'L1': 'L'}, 
            []),
    }
    component_structures = {
        'M': Component({'a', 'b'}),
        'L': Component({'a', 'b'}),
        'X': Component({'a'}),
    }

    grouped_ids = group_assemblies_by_isomorphism(id_to_graph, component_structures)

    assert grouped_ids == {
        0: {0, 1, 2},
        3: {3},
        4: {4},
    }


def test_group_assemblies_by_isomorphism_with_non_isomorphic_groups():
    id_to_graph = {
        0: Assembly(  # MLX
            {'M1': 'M', 'L1': 'L', 'X1': 'X'}, 
            [('M1.a', 'L1.a'), ('M1.b', 'X1.a')]),
        1: Assembly(  # MLX
            {'M2': 'M', 'L2': 'L', 'X2': 'X'}, 
            [('M2.a', 'L2.a'), ('M2.b', 'X2.a')]),
        2: Assembly(  # MLX
            {'M3': 'M', 'L3': 'L', 'X3': 'X'}, 
            [('M3.a', 'L3.a'), ('M3.b', 'X3.a')]),
        3: Assembly(  # MX2
            {'M1': 'M', 'X1': 'X', 'X2': 'X'}, 
            [('M1.a', 'X1.a'), ('M1.b', 'X2.a')]),
        4: Assembly(  # MX2
            {'M1': 'M', 'X1': 'X', 'X2': 'X'}, 
            [('M1.a', 'X1.a'), ('M1.b', 'X2.a')]),
        5: Assembly(  # L
            {'L1': 'L'}, 
            []),
        6: Assembly(  # L
            {'L1': 'L'}, 
            []),
    }
    component_structures = {
        'M': Component({'a', 'b'}),
        'L': Component({'a', 'b'}),
        'X': Component({'a'}),
    }
    non_isomorphic_groups = [{0, 1, 2}, {3, 4}]

    grouped_ids = group_assemblies_by_isomorphism(id_to_graph, component_structures, non_isomorphic_groups=non_isomorphic_groups)

    assert grouped_ids == {
        # Although 0, 1, 2 are isomorphic, since they are in the same
        # non-isomorphic group, their isomorphism check is skipped, thus
        # they are grouped to different isomorphic groups.
        0: {0},
        1: {1},
        2: {2},
        # The same applies to 3, 4.
        3: {3},
        4: {4},
        # 5, 6 are not in any non-isomorphic group, so they are checked
        # for isomorphism. They are isomorphic, so they are grouped together.
        5: {5, 6},
    }


if __name__ == '__main__':
    pytest.main(['-vv', __file__])



import pytest

from nasap_net.models import Assembly, AuxEdge, BindingSite, Bond, Component
from nasap_net.reaction_exploration_im.explorer import IntraReactionExplorer
from nasap_net.reaction_exploration_im.models import MLE, MLEKind, Reaction


@pytest.fixture
def M() -> Component:
    return Component(
        kind='M', sites=[0, 1, 2, 3],
        aux_edges=[AuxEdge(0, 1), AuxEdge(1, 2), AuxEdge(2, 3), AuxEdge(3, 0)])

@pytest.fixture
def L() -> Component:
    return Component(kind='L', sites=[0, 1])

@pytest.fixture
def X() -> Component:
    return Component(kind='X', sites=[0])

@pytest.fixture
def ML2X2_cis(M, L, X) -> Assembly:
    return Assembly(
        id_='ML2X2_cis',
        components={'M0': M, 'L0': L, 'L1': L, 'X0': X, 'X1': X},
        bonds=[
            Bond('M0', 0, 'X0', 0),
            Bond('M0', 1, 'X1', 0),
            Bond('M0', 2, 'L0', 0),
            Bond('M0', 3, 'L1', 0),
        ]
    )

@pytest.fixture
def ML2X_trans_ring(M, L, X) -> Assembly:
    return Assembly(
        components={'M0': M, 'L0': L, 'L1': L, 'X1': X},
        bonds=[
            Bond('M0', 0, 'L0', 1),
            Bond('M0', 1, 'X1', 0),
            Bond('M0', 2, 'L0', 0),
            Bond('M0', 3, 'L1', 0),
        ]
    )

@pytest.fixture
def ML2X_cis_ring(M, L, X) -> Assembly:
    return Assembly(
        components={'M0': M, 'L0': L, 'L1': L, 'X1': X},
        bonds=[
            Bond('M0', 0, 'L1', 1),
            Bond('M0', 1, 'X1', 0),
            Bond('M0', 2, 'L0', 0),
            Bond('M0', 3, 'L1', 0)
        ]
    )

@pytest.fixture
def free_X(X) -> Assembly:
    return Assembly(
        components={'X0': X},
        bonds=[]
    )


def test_explore(ML2X2_cis, ML2X_trans_ring, ML2X_cis_ring, free_X):
    explorer = IntraReactionExplorer(ML2X2_cis, MLEKind('M', 'X', 'L'))
    assert set(explorer.explore()) == {
        Reaction(
            init_assem=ML2X2_cis,
            entering_assem=None,
            product_assem=ML2X_trans_ring,
            leaving_assem=free_X,
            metal_bs=BindingSite('M0', 0),
            leaving_bs=BindingSite('X0', 0),
            entering_bs=BindingSite('L0', 1),
            duplicate_count=2
        ),
        Reaction(
            init_assem=ML2X2_cis,
            entering_assem=None,
            product_assem=ML2X_cis_ring,
            leaving_assem=free_X,
            metal_bs=BindingSite('M0', 0),
            leaving_bs=BindingSite('X0', 0),
            entering_bs=BindingSite('L1', 1),
            duplicate_count=2
        )
    }


def test__iter_mles(ML2X2_cis):
    explorer = IntraReactionExplorer(ML2X2_cis, MLEKind('M', 'X', 'L'))
    mles = set(explorer._iter_mles())
    assert mles == {
        MLE(BindingSite('M0', 0), BindingSite('X0', 0), BindingSite('L0', 1)),
        MLE(BindingSite('M0', 0), BindingSite('X0', 0), BindingSite('L1', 1)),
        MLE(BindingSite('M0', 1), BindingSite('X1', 0), BindingSite('L0', 1)),
        MLE(BindingSite('M0', 1), BindingSite('X1', 0), BindingSite('L1', 1)),
    }


def test__get_unique_mles(ML2X2_cis):
    explorer = IntraReactionExplorer(ML2X2_cis, MLEKind('M', 'X', 'L'))
    mles = {
        MLE(BindingSite('M0', 0), BindingSite('X0', 0), BindingSite('L0', 1)),
        MLE(BindingSite('M0', 0), BindingSite('X0', 0), BindingSite('L1', 1)),
        MLE(BindingSite('M0', 1), BindingSite('X1', 0), BindingSite('L0', 1)),
        MLE(BindingSite('M0', 1), BindingSite('X1', 0), BindingSite('L1', 1)),
    }
    assert set(explorer._get_unique_mles(mles)) == {
        MLE(
            BindingSite('M0', 0), BindingSite('X0', 0), BindingSite('L0', 1),
            duplication=2),
        MLE(
            BindingSite('M0', 0), BindingSite('X0', 0), BindingSite('L1', 1),
            duplication=2),
    }


def test__perform_reaction(ML2X2_cis, ML2X_trans_ring, free_X):
    explorer = IntraReactionExplorer(ML2X2_cis, MLEKind('M', 'X', 'L'))
    mle_with_dup = MLE(
        BindingSite('M0', 0), BindingSite('X0', 0), BindingSite('L0', 1),
        duplication=2)

    assert explorer._perform_reaction(mle_with_dup) == Reaction(
        init_assem=ML2X2_cis,
        entering_assem=None,
        product_assem=ML2X_trans_ring,
        leaving_assem=free_X,
        metal_bs=BindingSite('M0', 0),
        leaving_bs=BindingSite('X0', 0),
        entering_bs=BindingSite('L0', 1),
        duplicate_count=2
    )



from dataclasses import dataclass, field
from typing import Iterable

from nasap_net.isomorphism import get_all_isomorphisms
from nasap_net.models import Assembly, BindingSite
from nasap_net.utils import UnionFind


@dataclass(frozen=True)
class UniqueComb:
    """A unique binding site or binding site set with duplication count."""
    site_comb: tuple[BindingSite, ...]
    duplication: int = field(kw_only=True)


def extract_unique_site_combinations(
        binding_site_combs: Iterable[tuple[BindingSite, ...]],
        assembly: Assembly,
        ) -> set[UniqueComb]:
    """Compute unique binding sites or binding site sets."""
    grouped_node_combs = group_equivalent_node_combs(
        binding_site_combs, assembly)

    return {
        UniqueComb(
            site_comb=sorted(comb_group)[0],
            duplication=len(comb_group))
        for comb_group in grouped_node_combs
    }


def group_equivalent_node_combs(
        node_combs: Iterable[tuple[BindingSite, ...]],
        assembly: Assembly,
        ) -> set[frozenset[tuple[BindingSite, ...]]]:
    """Group equivalent node combinations."""
    node_combs = set(node_combs)
    uf = UnionFind(node_combs)

    self_isomorphisms = get_all_isomorphisms(assembly, assembly)
    binding_site_isoms = [
        isom.binding_site_mapping for isom in self_isomorphisms]

    for isom in binding_site_isoms:
        for comb in node_combs:
            mapped_comb = tuple(isom[site] for site in comb)
            if mapped_comb in node_combs:
                uf.union(comb, mapped_comb)
    return {
        frozenset(elements) for elements
        in uf.root_to_elements.values()}



import os
from collections.abc import Mapping, Sequence
from typing import Any, Literal, overload

from nasap_net import Assembly, Component, group_assemblies_by_isomorphism
from nasap_net.pipelines.lib import read_file, write_output


@overload
def concatenate_assemblies_pipeline(
        assemblies_path_list: Sequence[os.PathLike | str],
        components_path: None,
        resulting_assems_path: os.PathLike | str,
        *,
        skip_isomorphism_checks: Literal[True] = True,
        start: int = 0,
        overwrite: bool = False,
        verbose: bool = False,
        ) -> None: ...
@overload
def concatenate_assemblies_pipeline(
        assemblies_path_list: Sequence[os.PathLike | str],
        components_path: os.PathLike | str,
        resulting_assems_path: os.PathLike | str,
        *,
        already_unique_within_files: bool = False,
        skip_isomorphism_checks: Literal[False] = False,
        start: int = 0,
        overwrite: bool = False,
        verbose: bool = False,
        ) -> None: ...
def concatenate_assemblies_pipeline(
        assemblies_path_list: Sequence[os.PathLike | str],
        components_path: os.PathLike | str | None,
        resulting_assems_path: os.PathLike | str,
        *,
        already_unique_within_files: bool = False,
        skip_isomorphism_checks: bool = False,
        start: int = 0,
        overwrite: bool = False,
        verbose: bool = False,
        ) -> None:
    """Concatenate a list of assemblies into a single list excluding duplicates.
    
    Concatenates a list of assemblies from different files into a single 
    list. The IDs of the assemblies are reindexed starting from the given
    start value.

    Parameters
    ----------
    assemblies_path_list : Sequence[os.PathLike | str]
        List of paths to the files containing the assemblies.
        Each file should contain a dictionary with the assembly IDs as keys
        and the Assembly objects as values.
    components_path : os.PathLike | str | None
        Path to the file containing the component kinds. 
        None is allowed only if `skip_isomorphism_checks` is True.
    resulting_assems_path : os.PathLike | str
        Path to the file to save the concatenated list of assemblies.
    already_unique_within_files : bool, optional
        Whether the assemblies in each file are already unique, 
        by default False. If True, the isomorphism checks are skipped 
        for the assemblies within each file. The parameter will be 
        ignored if `exclude_duplicates` is False.
    skip_isomorphism_checks : bool, optional
        Whether to skip all isomorphism checks, by default False.
        If True, the assemblies are concatenated without checking for
        isomorphism. The resulting list may contain duplicate assemblies.
    start : int, optional
        Starting index for the reindexing of the assemblies, by default 0.
    overwrite : bool, optional
        Whether to overwrite the file if it already exists, by default False.
    verbose : bool, optional
        Whether to print the process steps, by default False
    
    Notes
    -----
    The reindexing order is determined by the order of the input files and 
    the order of the assemblies in each file, not by their alphabetical 
    order.

    Output file contains a dictionary with the reindexed IDs as keys and the
    Assembly objects as values.
    """
    if skip_isomorphism_checks:
        concatenate_assemblies_without_isom_checks(
            assemblies_path_list, resulting_assems_path,
            start=start, overwrite=overwrite, verbose=verbose)
        return
    if components_path is None:
        raise ValueError(
            'The parameter `components_path` is required if '
            '`skip_isomorphism_checks` is False.')
    
    # Input
    list_of_id_to_assembly: Sequence[Mapping[Any, Assembly]] = [
        read_file(path, verbose=verbose)
        for path in assemblies_path_list]
    
    components: Mapping[str, Component] = read_file(
        components_path, verbose=verbose)['component_kinds']
    
    # Main process
    if verbose:
        print('Concatenating assembly lists...')

    assemblies: list[Assembly] = []
    for id_to_assembly in list_of_id_to_assembly:
        assemblies.extend(id_to_assembly.values())
    
    reindexed_with_dup = {
        new_id: assembly for new_id, assembly 
        in enumerate(assemblies, start=start)
    }

    if verbose:
        print('Concatenation completed.')
    if verbose:
        print('Excluding duplicate assemblies...')
    
    if already_unique_within_files:
        non_isomorphic_groups = [
            set(id_to_assembly.keys())
            for id_to_assembly in list_of_id_to_assembly
        ]
    else:
        non_isomorphic_groups = None
    
    isom_group = group_assemblies_by_isomorphism(
        reindexed_with_dup, components,
        non_isomorphic_groups=non_isomorphic_groups)
    
    reindexed_without_dup = {
        id_: reindexed_with_dup[id_] for id_ in isom_group.keys()}
    
    if verbose:
        print('Duplicate exclusion completed.')

    # Save reindexed assemblies
    write_output(
        resulting_assems_path, reindexed_without_dup,
        overwrite=overwrite, verbose=verbose,
        header='Concatenated list of assemblies')


def concatenate_assemblies_without_isom_checks(
        assemblies_path_list: Sequence[os.PathLike | str],
        resulting_assems_path: os.PathLike | str,
        *,
        start: int = 0,
        overwrite: bool = False,
        verbose: bool = False,
        ) -> None:
    """Concatenate a list of assemblies into a single list.
    
    Concatenates a list of assemblies from different files into a single 
    list. The IDs of the assemblies are reindexed starting from the given
    start value.

    Parameters
    ----------
    assemblies_path_list : Sequence[os.PathLike | str]
        List of paths to the files containing the assemblies.
        Each file should contain a dictionary with the assembly IDs as keys
        and the Assembly objects as values.
    resulting_assems_path : os.PathLike | str
        Path to the file to save the concatenated list of assemblies.
    exclude_duplicates : bool, optional
        Whether to exclude duplicate assemblies, by default False.
        If True, isomorphism checks are performed to group the assemblies
        by isomorphism. The resulting list contains only the unique
        assemblies.
    already_unique_within_files : bool, optional
        Whether the assemblies in each file are already unique, 
        by default False. If True, the isomorphism checks are skipped 
        for the assemblies within each file. The parameter will be 
        ignored if `exclude_duplicates` is False.
    start : int, optional
        Starting index for the reindexing of the assemblies, by default 0.
    overwrite : bool, optional
        Whether to overwrite the file if it already exists, by default False.
    verbose : bool, optional
        Whether to print the process steps, by default False
    
    Notes
    -----
    The reindexing order is determined by the order of the input files and 
    the order of the assemblies in each file, not by their alphabetical 
    order.

    Output file contains a dictionary with the reindexed IDs as keys and the
    Assembly objects as values.
    """
    # Input
    list_of_id_to_assembly: Sequence[Mapping[Any, Assembly]] = [
        read_file(path, verbose=verbose)
        for path in assemblies_path_list]
    
    # Main process
    if verbose:
        print('Concatenating assembly lists...')

    assemblies: list[Assembly] = []
    for id_to_assembly in list_of_id_to_assembly:
        assemblies.extend(id_to_assembly.values())
    
    reindexed = {
        new_id: assembly for new_id, assembly 
        in enumerate(assemblies, start=start)
    }

    if verbose:
        print('Concatenation completed.')

    # Save reindexed assemblies
    write_output(
        resulting_assems_path, reindexed,
        overwrite=overwrite, verbose=verbose,
        header='Concatenated list of assemblies')



from __future__ import annotations

from collections.abc import Callable, Iterable, Iterator, Mapping
from copy import copy, deepcopy
from dataclasses import dataclass
from functools import wraps
from itertools import chain
from types import MappingProxyType
from typing import (Concatenate, Literal, ParamSpec, TypeVar, TypedDict,
                    overload)

import networkx as nx
import yaml

from .bindsite_id_converter import BindsiteIdConverter
from .component import Component

__all__ = ['Assembly', 'assembly_to_graph', 'assembly_to_rough_graph', 'find_free_bindsites']


# For type hint of the decorator
P = ParamSpec("P")
R = TypeVar("R")


@dataclass
class AbsAuxEdge:
    """An auxiliary edge specified by absolute binding sites."""
    bindsite1: str
    bindsite2: str
    aux_type: str


class Assembly(yaml.YAMLObject):
    """A class to represent an assembly.
    
    An assembly is a group of components connected by bonds.
    """
    yaml_loader = yaml.SafeLoader
    yaml_dumper = yaml.Dumper
    yaml_tag = '!Assembly'
    yaml_flow_style = None
    
    def __init__(
            self, 
            comp_id_to_kind: Mapping[str, str] | None = None,
            bonds: (
                Iterable[tuple[str, str]] | Iterable[Iterable[str]] | None
                ) = None,
            name: str | None = None
            ) -> None:
        self._comp_id_to_kind = dict[str, str]()
        self._bonds = set[frozenset[str]]()
        self.name = name

        if comp_id_to_kind is not None:
            for component_id, component_kind in comp_id_to_kind.items():
                self.add_component(component_id, component_kind)
        if bonds is not None:
            for bindsite1, bindsite2 in bonds:
                self.add_bond(bindsite1, bindsite2)
            
        self._rough_g_cache: nx.Graph | None = None
        self._bindsite_to_connected_cache: dict[str, str] | None = None
    
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Assembly):
            return False
        return self.comp_id_to_kind == other.comp_id_to_kind and\
            self.bonds == other.bonds and self.name == other.name

    # Decorator
    # For type hint of the decorator, see the following link:
    # https://github.com/microsoft/pyright/issues/6472
    @staticmethod
    def _clear_g_caches(func: Callable[Concatenate[Assembly, P], R]
            ) -> Callable[Concatenate[Assembly, P], R]:
        """Decorator to clear the cache of the graph snapshot before
        calling the method."""
        @wraps(func)
        def wrapper(self: Assembly, *args: P.args, **kwargs: P.kwargs):
            self._rough_g_cache = None
            self._bindsite_to_connected_cache = None
            return func(self, *args, **kwargs)
        return wrapper

    # ============================================================
    # Properties (read-only)
    # ============================================================

    @property
    def comp_id_to_kind(self) -> MappingProxyType[str, str]:
        """Return a read-only view of the components.
        
        The returned object can be used like a dictionary, but it is
        read-only. Changes to the original assembly will be reflected
        in the returned object.
        """
        return MappingProxyType(self._comp_id_to_kind)
    
    @property
    def component_ids(self) -> set[str]:
        return set(self._comp_id_to_kind.keys())

    @property
    def component_kinds(self) -> set[str]:
        return set(self._comp_id_to_kind.values())
    
    @property
    def bonds(self) -> set[frozenset[str]]:
        return self._bonds.copy()
    
    @property
    def bindsite_to_connected(self) -> dict[str, str]:
        if getattr(self, '_bindsite_to_connected', None) is None:
            self._bindsite_to_connected_cache =\
                self.create_bindsite_to_connected()
        assert self._bindsite_to_connected_cache is not None
        return copy(self._bindsite_to_connected_cache)
    
    def create_bindsite_to_connected(self) -> dict[str, str]:
        d = {}
        for bond in self._bonds:
            bindsite1, bindsite2 = bond
            d[bindsite1] = bindsite2
            d[bindsite2] = bindsite1
        return d
    
    def g_snapshot(
            self, component_structures: Mapping[str, Component]
            ) -> nx.Graph:
        """Returns a snapshot of the assembly graph.
        
        The snapshot is a deep copy of the assembly graph. Therefore,
        any modification to the snapshot will not affect the assembly.
        """
        # Prevent the user from modifying the assembly.
        return self._to_graph(component_structures)
    
    @property
    def rough_g_snapshot(self) -> nx.Graph:
        """Returns a rough graph of the assembly."""
        if getattr(self, '_rough_g_cache', None) is None:
            self._rough_g_cache = self._to_rough_graph()
        return deepcopy(self._rough_g_cache)
    
    # ============================================================
    # Methods to modify the assembly (using relative names)
    # ============================================================
    
    @_clear_g_caches
    def add_component(
            self, component_id: str, component_kind: str) -> None:
        """Add a component to the assembly.
        
        Note: No bond is added between the component and the assembly.
        The user should add bonds between the component and the assembly
        if necessary.
        """
        self._comp_id_to_kind[component_id] = component_kind

    @_clear_g_caches
    def remove_component(self, component_id: str) -> None:
        del self._comp_id_to_kind[component_id]
    
    @_clear_g_caches
    def add_bond(self, bindsite1: str, bindsite2: str) -> None:
        """Add a bond to the assembly."""
        id_converter = BindsiteIdConverter()
        comp1, rel1 = id_converter.global_to_local(bindsite1)
        comp2, rel2 = id_converter.global_to_local(bindsite2)
        for comp in [comp1, comp2]:
            if comp not in self._comp_id_to_kind:
                raise ValueError(
                    f'The component "{comp}" does not exist in the assembly.')
        self._bonds.add(frozenset([bindsite1, bindsite2]))

    @_clear_g_caches
    def remove_bond(
            self, bindsite1: str, bindsite2: str) -> None:
        """Remove a bond from the assembly."""
        self._bonds.remove(frozenset([bindsite1, bindsite2]))

    # ============================================================
    # Methods to make multiple modifications at once
    # ============================================================

    def add_components(self, components: Iterable[tuple[str, str]]) -> None:
        for component_id, component_kind in components:
            self.add_component(component_id, component_kind)

    def remove_components(self, component_ids: Iterable[str]) -> None:
        for component_id in component_ids:
            self.remove_component(component_id)

    def add_bonds(
            self, bonds: Iterable[tuple[str, str]]) -> None:
        for bindsite1, bindsite2 in bonds:
            self.add_bond(bindsite1, bindsite2)

    def remove_bonds(
            self, bonds: Iterable[tuple[str, str]]) -> None:
        for bindsite1, bindsite2 in bonds:
            self.remove_bond(bindsite1, bindsite2)
    
    # ============================================================
    # Methods to relabel the assembly
    # ============================================================
    
    # `@overload` decorator is just for type hinting;
    # it does not affect the behavior of the method.
    @overload
    @_clear_g_caches
    def rename_component_ids(
            self, mapping: Mapping[str, str],
            *, copy: Literal[True] = True) -> Assembly:
        ...
    @overload
    @_clear_g_caches
    def rename_component_ids(
            self, mapping: Mapping[str, str],
            *, copy: Literal[False]) -> None:
        ...
    @_clear_g_caches
    def rename_component_ids(
            self, mapping: Mapping[str, str],
            *, copy: bool = True) -> Assembly | None:
        if copy:
            assem = deepcopy(self)
        else:
            assem = self

        id_converter = BindsiteIdConverter()

        new_components = {}
        for old_id, component in assem._comp_id_to_kind.items():
            new_id = mapping.get(old_id, old_id)
            new_components[new_id] = component
        
        new_bonds = set()
        for bindsite1, bindsite2 in assem._bonds:
            comp1, rel1 = id_converter.global_to_local(bindsite1)
            comp2, rel2 = id_converter.global_to_local(bindsite2)
            new_bindsite1 = id_converter.local_to_global(
                mapping.get(comp1, comp1), rel1)
            new_bindsite2 = id_converter.local_to_global(
                mapping.get(comp2, comp2), rel2)
            new_bonds.add(frozenset([new_bindsite1, new_bindsite2]))

        assem._comp_id_to_kind = new_components
        assem._bonds = new_bonds
        assem._rough_g_cache = None

        # TODO: Check if the new component IDs are valid.
        
        return assem

    # ============================================================
    # Helper methods
    # ============================================================

    @overload
    def get_connected_bindsite(
            self, bindsite: str, error_if_free: Literal[False] = False
            ) -> str | None: ...
    @overload
    def get_connected_bindsite(
            self, bindsite: str, error_if_free: Literal[True]
            ) -> str: ...
    def get_connected_bindsite(
            self, bindsite, error_if_free=False):
        """Get the connected binding site of the binding site.
        
        Parameters
        ----------
        bindsite : str
            The binding site.
        error_if_free : bool, optional
            If True, raise an error if the binding site is free.
            It is useful when the user expects the binding site to be
            connected.

        Returns
        -------
        str | None
            The connected binding site. If the binding site is free,
            and `error_if_free` is False, return None.
        """
        connected = self.bindsite_to_connected.get(bindsite)
        if connected is None and error_if_free:
            raise ValueError(
                f'The binding site "{bindsite}" is free.')
        return connected
    
    def is_free_bindsite(self, bindsite: str) -> bool:
        return self.get_connected_bindsite(bindsite) is None
    
    def get_component_kind(self, component_id: str) -> str:
        return self._comp_id_to_kind[component_id]
    
    def get_component_kind_of_core(self, core: str) -> str:
        id_converter = BindsiteIdConverter()
        comp_id, rel = id_converter.global_to_local(core)
        return self.get_component_kind(comp_id)
    
    def get_component_kind_of_bindsite(self, bindsite: str) -> str:
        id_converter = BindsiteIdConverter()
        comp_id, rel = id_converter.global_to_local(bindsite)
        return self.get_component_kind(comp_id)
    
    def deepcopy(self) -> Assembly:
        return deepcopy(self)
    
    def get_core_of_the_component(self, component_id: str) -> str:
        id_converter = BindsiteIdConverter()
        return id_converter.local_to_global(component_id, 'core')

    def iter_all_cores(self) -> Iterator[str]:
        id_converter = BindsiteIdConverter()
        for comp_id, comp_kind in self.comp_id_to_kind.items():
            yield id_converter.local_to_global(comp_id, 'core')
    
    def get_all_bindsites(
            self, component_structures: Mapping[str, Component]
            ) -> set[str]:
        """Get all the binding sites in the assembly."""
        # TODO: Consider yielding the binding sites instead of returning a set.
        id_converter = BindsiteIdConverter()
        all_bindsites = set()
        for comp_id, comp_kind in self.comp_id_to_kind.items():
            comp_struct = component_structures[comp_kind]
            for bindsite in comp_struct.binding_sites:
                all_bindsites.add(id_converter.local_to_global(comp_id, bindsite))
        return all_bindsites

    def iter_aux_edges(
            self, component_structures: Mapping[str, Component]
            ) -> Iterator[AbsAuxEdge]:
        id_converter = BindsiteIdConverter()
        for comp_id, comp_kind in self.comp_id_to_kind.items():
            comp_struct = component_structures[comp_kind]
            for rel_aux_edge in comp_struct.aux_edges:
                yield AbsAuxEdge(
                    id_converter.local_to_global(
                        comp_id, rel_aux_edge.local_binding_site1),
                    id_converter.local_to_global(
                        comp_id, rel_aux_edge.local_binding_site2),
                    rel_aux_edge.aux_kind)

    def get_bindsites_of_component(
            self, component_id: str, 
            component_structures: Mapping[str, Component]
            ) -> set[str]:
        """Get the binding sites of the component."""
        id_converter = BindsiteIdConverter()
        comp_kind = self.get_component_kind(component_id)
        comp_struct = component_structures[comp_kind]
        return {
            id_converter.local_to_global(component_id, bindsite)
            for bindsite in comp_struct.binding_sites}
    
    def get_all_bindsites_of_kind(
            self, component_kind: str,
            component_structures: Mapping[str, Component],
            ) -> Iterator[str]:
        for comp_id, comp in self.comp_id_to_kind.items():
            if comp == component_kind:
                yield from self.get_bindsites_of_component(
                    comp_id, component_structures)

    def find_free_bindsites(
            self, component_structures: Mapping[str, Component]
            ) -> set[str]:
        """Find free bindsites."""
        id_converter = BindsiteIdConverter()
        all_bindsites = {
            id_converter.local_to_global(comp_id, bindsite)
            for comp_id, comp_kind in self.comp_id_to_kind.items()
            for bindsite in component_structures[comp_kind].binding_sites
        }
        connected_bindsites = chain(*self.bonds)
        free_bindsites = all_bindsites - set(connected_bindsites)
        return free_bindsites

    # ============================================================
    # Methods to convert the assembly to graphs
    # ============================================================

    def _to_graph(
            self, component_kinds: Mapping[str, Component]
            ) -> nx.Graph:
        return assembly_to_graph(self, component_kinds)

    def _to_rough_graph(self) -> nx.Graph:
        id_converter = BindsiteIdConverter()
        G = nx.MultiGraph()
        for comp_id, comp_kind in self.comp_id_to_kind.items():
            G.add_node(comp_id, component_kind=comp_kind)
        for bindsite1, bindsite2 in self.bonds:
            comp1, rel1 = id_converter.global_to_local(bindsite1)
            comp2, rel2 = id_converter.global_to_local(bindsite2)
            G.add_edge(comp1, comp2, bindsites={comp1: rel1, comp2: rel2})
        return G

    @classmethod
    def bond_to_rough_bond(
            cls, bond: Iterable[str]) -> list[str]:
        # TODO: Move this method to a separate module.
        id_converter = BindsiteIdConverter()
        bindsites = list(bond)
        comp1, rel1 = id_converter.global_to_local(bindsites[0])
        comp2, rel2 = id_converter.global_to_local(bindsites[1])
        return [comp1, comp2]
    
    @classmethod
    def from_yaml(cls, loader, node):
        data = loader.construct_mapping(node, deep=True)
        return cls(
            data.get('comp_id_to_kind', None),
            data.get('bonds', None),
            data.get('name', None))
    
    @classmethod
    def to_yaml(cls, dumper, data: Assembly):
        DataDict = TypedDict(
            'DataDict', {
                'comp_id_to_kind': dict[str, str],
                'bonds': list[list[str]],
                'name': str},
            total=False)
        data_dict: DataDict = {}
        if data.comp_id_to_kind:
            data_dict['comp_id_to_kind'] = dict(data.comp_id_to_kind)
        if data.bonds:
            data_dict['bonds'] = sorted(sorted(bond) for bond in data.bonds)
        if data.name:
            data_dict['name'] = data.name
        return dumper.represent_mapping(
            cls.yaml_tag, data_dict, flow_style=cls.yaml_flow_style)


def assembly_to_graph(
        assembly: Assembly,
        component_structures: Mapping[str, Component],
        ) -> nx.Graph:
    G = nx.Graph()
    for comp_id, comp_kind in assembly.comp_id_to_kind.items():
        comp_structure = component_structures[comp_kind]
        add_component_to_graph(G, comp_id, comp_kind, comp_structure)
    for bond in assembly.bonds:
        G.add_edge(*bond)
    return G


def add_component_to_graph(
        g: nx.Graph,
        component_id: str, component_kind: str,
        component_structure: Component,
        ) -> None:
    # Add the core node
    id_converter = BindsiteIdConverter()
    core_abs = id_converter.local_to_global(component_id, 'core')
    g.add_node(
        core_abs, core_or_bindsite='core', component_kind=component_kind)
    
    # Add the binding sites
    for bindsite in component_structure.binding_sites:
        bindsite_abs = id_converter.local_to_global(component_id, bindsite)
        g.add_node(bindsite_abs, core_or_bindsite='bindsite')
        g.add_edge(core_abs, bindsite_abs)
    
    # Add the auxiliary edges
    for aux_edge in component_structure.aux_edges:
        bs1_abs = id_converter.local_to_global(component_id, aux_edge.local_binding_site1)
        bs2_abs = id_converter.local_to_global(component_id, aux_edge.local_binding_site2)
        g.add_edge(bs1_abs, bs2_abs, aux_type=aux_edge.aux_kind)



from typing import Protocol, TypeAlias, TypeVar

S = TypeVar('S', int, str)  # Site ID
C = TypeVar('C', int, str)  # Component ID
A = TypeVar('A', int, str)  # Assembly ID
R = TypeVar('R', int, str)  # Reaction ID
ID: TypeAlias = int | str


class SupportsDunderLt(Protocol):
    def __lt__(self, other: object) -> bool:
        ...



import pytest

from nasap_net.models import Assembly, Bond, Component
from nasap_net.reaction_exploration_im.lib import AssemblyNotFoundError, \
    IsomorphicAssemblyFinder


@pytest.fixture
def M():
    return Component(kind='M', sites=[0, 1])

@pytest.fixture
def L():
    return Component(kind='L', sites=[0, 1])

@pytest.fixture
def X():
    return Component(kind='X', sites=[0])


def test_find(M, L, X):
    # X0(0)-(0)M0(1)-(0)L0(1)
    MLX = Assembly(
        components={'X0': X, 'M0': M, 'L0': L},
        bonds=[Bond('X0', 0, 'M0', 0), Bond('M0', 1, 'L0', 0)]
    )
    another_MLX = Assembly(
        components={'X1': X, 'M1': M, 'L1': L},
        bonds=[Bond('X1', 0, 'M1', 0), Bond('M1', 1, 'L1', 0)]
    )

    finder = IsomorphicAssemblyFinder(search_space=[another_MLX])
    result = finder.find(MLX)
    assert result == another_MLX


def test_not_found_error(M, L, X):
    # X0(0)-(0)M0(1)-(0)L0(1)
    MLX = Assembly(
        components={'X0': X, 'M0': M, 'L0': L},
        bonds=[Bond('X0', 0, 'M0', 0), Bond('M0', 1, 'L0', 0)]
    )
    # X0(0)-(0)M0(1)
    MX = Assembly(
        components={'X0': X, 'M0': M},
        bonds=[Bond('X0', 0, 'M0', 0)]
    )

    finder = IsomorphicAssemblyFinder(search_space=[MX])
    with pytest.raises(AssemblyNotFoundError):
        finder.find(MLX)


def test_duplicate_assemblies_not_raise_error(M, L, X):
    # X0(0)-(0)M0(1)-(0)L0(1)
    MLX = Assembly(
        components={'X0': X, 'M0': M, 'L0': L},
        bonds=[Bond('X0', 0, 'M0', 0), Bond('M0', 1, 'L0', 0)]
    )
    another_MLX = Assembly(
        components={'X1': X, 'M1': M, 'L1': L},
        bonds=[Bond('X1', 0, 'M1', 0), Bond('M1', 1, 'L1', 0)]
    )
    duplicate_MLX = Assembly(
        components={'X2': X, 'M2': M, 'L2': L},
        bonds=[Bond('X2', 0, 'M2', 0), Bond('M2', 1, 'L2', 0)]
    )

    finder = IsomorphicAssemblyFinder(
        search_space=[another_MLX, duplicate_MLX],
    )
    result = finder.find(MLX)
    assert result in {another_MLX, duplicate_MLX}
